# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import curses
import curses.ascii
import gzip
import json
import mmap
import shutil
import tempfile
import warnings
from argparse import ArgumentParser, BooleanOptionalAction
from array import array
from binascii import unhexlify
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum, unique
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO

import platformdirs
import zstandard as zstd

from gallia.log import PenlogPriority, PenlogRecord
from gallia.services.uds.core.service import NegativeResponse, UDSRequest, UDSResponse


@unique
class InterpretationColor(IntEnum):
    DEFAULT = 42
    UDS_REQUEST = 43
    UDS_POSITIVE_RESPONSE = 44
    UDS_NEGATIVE_RESPONSE = 45


@dataclass
class PenlogEntry(PenlogRecord):
    interpretation: str | None = None
    interpretation_color: InterpretationColor = InterpretationColor.DEFAULT


@dataclass
class FormattedText:
    text: str
    format: int

    @property
    def sanitized_text(self) -> str:
        # Strip 0 bytes…
        return self.text.encode().replace(b"\x00", b"").decode()


@dataclass
class DisplayEntry:
    """
    Represents one single line of a single formatted penlog entry as displayed in the console.

    :param texts: The (differently) formatted text fragments of the displayed line.
    :param penlog_entry_number: The line number of the corresponding penlog entry.
    :param entry_line_number: The index in the list of formatted lines as created for the corresponding penlog entry.
    :param last_line: True if and only if this entry refers to the last of the formatted lines as created for the
                      corresponding penlog entry.
    """

    texts: list[FormattedText]
    penlog_entry_number: int
    entry_line_number: int
    last_line = False


@dataclass
class PriorityZone:
    """
    Represents a continuous range (including start and end) of entries for which a certain priority filter applies.

    :param start: The first penlog entry of the zone.
    :param end: The last penlog entry of the zone.
    :param priority: The filtering priority to be applied on all entries in in the zone.
    """

    start: int
    end: int | None
    priority: PenlogPriority


# NOTE: Workaround for compatibility with Python 3.11 and older, see:
# https://github.com/python/mypy/issues/13942
if TYPE_CHECKING:
    IntArray = array[int]
else:
    IntArray = array


class EntryCache:
    """
    A simple two level cache that stores the latest accessed penlog entries.

    This is used to speed up the processing of penlog entries by reducing the number of necessary fetches from the
    input file. It may be ineffective if the whole file is kept in memory.
    """

    def __init__(
        self,
        file: BinaryIO | mmap.mmap,
        entry_positions: IntArray,
        cache_size: int = 20_000,
    ):
        self.file = file
        self.old_entries: dict[int, PenlogEntry] = {}
        self.entries: dict[int, PenlogEntry] = {}
        self.entry_positions = entry_positions
        self.cache_size = cache_size

    def __len__(self) -> int:
        return len(self.entry_positions)

    def __getitem__(self, entry_number: int) -> PenlogEntry:
        try:
            return self.entries[entry_number]
        except KeyError:
            self._load_entry(entry_number)
            return self.entries[entry_number]

    def _load_entry(self, entry_number: int) -> None:
        if len(self.entries) + len(self.old_entries) >= self.cache_size:
            self.old_entries = self.entries
            self.entries = {}

        try:
            self.entries[entry_number] = self.old_entries.pop(entry_number)
        except KeyError:
            self.file.seek(self.entry_positions[entry_number])
            line = self.file.readline()

            try:
                self.entries[entry_number] = PenlogEntry.parse_json(line)
            except (json.decoder.JSONDecodeError, TypeError):
                self.entries[entry_number] = PenlogEntry(
                    data=line.decode("utf-8"),
                    host="",
                    module="JSON",
                    datetime=datetime.fromtimestamp(0),
                    tags=["ERROR"],
                    priority=PenlogPriority.ERROR,
                )


@dataclass
class Configuration:
    priority_zones: list[PriorityZone]
    filter: list[str]
    interpret: bool


class CursedHR:
    """
    A curses based interactive reader for log files complying to penlog.

    It supports the following features:
        - interactive display of penlog files (similar to less)
        - coloring and formatting of entries based on the priority level
        - optimized sectional filtering based on priority
        - unoptimized sectional filtering based on any penlog entry attribute
        - in-line interpretation of UDS messages using the Gallia UDS implementation

    See https://github.com/Fraunhofer-AISEC/penlog for the penlog specification.
    """

    def __init__(
        self,
        in_file: Path,
        priority: PenlogPriority = PenlogPriority.DEBUG,
        filters: list[str] | None = None,
        use_prefix: bool = True,
        relative_timings: bool = False,
    ):
        self.in_file = in_file
        self.level_pointers: list[array[int]] = [array("L") for _ in range(9)]
        self.entry_positions = array("L")
        self.configuration_history = [
            Configuration(
                priority_zones=[PriorityZone(0, None, priority)],
                filter=filters if filters is not None else [],
                interpret=False,
            )
        ]
        self.configuration_index = 0
        self.use_prefix = use_prefix
        self.relative_timings = relative_timings

        self.priority_keys = {
            "m": PenlogPriority.EMERGENCY,
            "a": PenlogPriority.ALERT,
            "c": PenlogPriority.CRITICAL,
            "e": PenlogPriority.ERROR,
            "w": PenlogPriority.WARNING,
            "n": PenlogPriority.NOTICE,
            "i": PenlogPriority.INFO,
            "d": PenlogPriority.DEBUG,
            "t": PenlogPriority.TRACE,
        }

        try:
            self.window = self.init_curses()
            self.color_ids = self.define_colors()
            self.file = self.uncompressed_file()

            try:
                with mmap.mmap(self.file.fileno(), 0, access=mmap.ACCESS_READ) as mm_file:
                    self.parse_structure(mm_file)
                    self.entries = EntryCache(mm_file, self.entry_positions)
                    self.reference_time = self.entries[0].datetime
                    self.handle_io()
            finally:
                self.file.close()
        finally:
            self.terminate_curses()

    def init_curses(self) -> Any:
        window = curses.initscr()
        curses.start_color()
        curses.use_default_colors()
        curses.noecho()
        curses.cbreak()
        window.keypad(True)

        return window

    def terminate_curses(self) -> None:
        curses.nocbreak()
        self.window.keypad(False)
        curses.echo()
        curses.endwin()

    def define_colors(self) -> dict[int, int]:
        prio_colors = {
            PenlogPriority.EMERGENCY: (100, curses.COLOR_RED),
            PenlogPriority.ALERT: (101, curses.COLOR_RED),
            PenlogPriority.CRITICAL: (102, curses.COLOR_RED),
            PenlogPriority.ERROR: (103, curses.COLOR_RED),
            PenlogPriority.WARNING: (104, curses.COLOR_YELLOW),
            PenlogPriority.NOTICE: (105, -1),
            PenlogPriority.INFO: (106, -1),
            PenlogPriority.DEBUG: (107, 8),
            PenlogPriority.TRACE: (108, curses.COLOR_BLUE),
        }

        for identifier, value in prio_colors.values():
            curses.init_pair(identifier, value, -1)

        curses.init_pair(InterpretationColor.DEFAULT, 8, -1)
        curses.init_pair(InterpretationColor.UDS_REQUEST, curses.COLOR_CYAN, -1)
        curses.init_pair(InterpretationColor.UDS_POSITIVE_RESPONSE, curses.COLOR_GREEN, -1)
        curses.init_pair(InterpretationColor.UDS_NEGATIVE_RESPONSE, 166, -1)

        return {prio: color[0] for prio, color in prio_colors.items()}

    def uncompressed_file(self) -> BinaryIO:
        """
        Returns an uncompressed version of the input file as referenced by self.in_file.

        In case the input file is already uncompressed, the original file is returned.
        In any other case, the input file is decompressed to a temporary file (potentially on disk).
        """
        self.window.clear()
        self.window.addstr(f"Loading contents from {self.in_file}")
        self.window.refresh()

        if self.in_file.suffix in [".zst", ".gz"]:

            def copy_to_file(tmp_file: Any) -> None:
                match self.in_file.suffix:
                    case ".zst":
                        with zstd.open(self.in_file, "rb") as in_file:
                            shutil.copyfileobj(in_file, file)
                    case ".gz":
                        with gzip.open(self.in_file, "rb") as in_file:
                            shutil.copyfileobj(in_file, file)

                tmp_file.flush()

            self.window.erase()
            self.window.addstr(f"Loading contents from {self.in_file}: Decompressing file ...")
            self.window.refresh()

            file = tempfile.TemporaryFile()

            try:
                try:
                    copy_to_file(file)
                except OSError:
                    file.close()

                    self.window.erase()
                    self.window.addstr(
                        f"Could not decompress to {tempfile.gettempdir()}. Trying to decompress to source directory ..."
                    )
                    self.window.refresh()

                    file = tempfile.TemporaryFile(dir=platformdirs.user_cache_dir())

                    copy_to_file(file)
            except:
                file.close()
                raise
        else:
            file = self.in_file.open("rb")  # type: ignore

        return file

    def parse_structure(self, file: BinaryIO | mmap.mmap) -> None:
        """
        Parses an (already uncompressed) penlog file to a skeleton containing structural information but no data.

        This process enables handling relatively large penlog entries efficiently by extracting information on the
        priority from each entry, which can later be used to quickly jump to adjacent entries of at least same priority.
        This is accomplished by the following two data structures:

            - self.entry_positions: Stores the absolute offsets of each penlog entry inside the input file.
                                    This offset can later be used to efficiently retrieve the data for a penlog entry
                                    from the input file without the need to store it in memory.
            - self.level_pointers: Stores the entries which have at least a certain priority (i.e. less or equal).
                                   More precisely, the indices of the corresponding entries in self.entry_positions.
                                   There exists on such list for each priority level.
                                   This allows to quickly find entries which are to be displayed inside a priority zone
                                   even if there are many entries with filtered out priority in between.
        """
        prio_prefix = b'"priority":'
        prio_prefix_len = len(prio_prefix)

        file.seek(0, 2)
        file_length = file.tell()
        file.seek(0)

        self.window.erase()
        self.window.addstr(f"Loading contents from {self.in_file}: Parsing structure ({0}%)")
        self.window.refresh()

        prev_progress = 0
        num_entries = 0

        while True:
            n = file.tell()
            line = file.readline()

            if not line:
                break

            prio = PenlogPriority.INFO.value

            if line[0] == 60:  # <P> priority prefix
                prio = line[1] - 48
            elif line[0] != 123:  # no json
                prio = PenlogPriority.ERROR
            elif (pos := line.find(prio_prefix)) > 0:  # json with priority
                pos += prio_prefix_len

                while not 48 <= line[pos] <= 57:  # find first digit after key
                    pos += 1

                prio = line[pos] - 48

            for i in reversed(range(prio, PenlogPriority.TRACE + 1)):
                self.level_pointers[i].append(num_entries)

            self.entry_positions.append(n)
            num_entries += 1

            progress = n * 100 // file_length

            if progress >= prev_progress + 10:
                prev_progress = progress
                self.window.erase()
                self.window.addstr(
                    f"Loading contents from {self.in_file}: Parsing structure ({progress}%)"
                )
                self.window.refresh()

    @property
    def configuration(self) -> Configuration:
        return self.configuration_history[self.configuration_index]

    def new_configuration(self) -> None:
        self.configuration_index += 1
        self.configuration_history = self.configuration_history[: self.configuration_index]
        self.configuration_history.append(deepcopy(self.configuration_history[-1]))

    def entry_zone(self, entry_id: int) -> PriorityZone:
        """
        Returns the zone into which the entry with the given id falls.

        :params entry_id: The index of the entry.
        :return: The zone into which the entry with the given id falls.
        """
        for zone in self.configuration.priority_zones:
            if entry_id >= zone.start and (zone.end is None or entry_id <= zone.end):
                return zone

        raise AssertionError

    def update_zones(self, new_zone: PriorityZone) -> None:
        """
        Inserts the given zone into the current set of zones.

        Adds a new list of zones into the zone history, which contains the previous zones and the new zone.
        In the process of integrating the new zone, itself as well as old zones may be altered or completely removed.
        It is guaranteed, that the zones are always ordered and do not overlap.

        :param new_zone: The zone to be added.
        """
        assert new_zone.end is not None

        self.new_configuration()

        i = 0

        while i < len(self.configuration.priority_zones) - 1:
            cur_zone = self.configuration.priority_zones[i]
            assert cur_zone.end is not None

            # If no intersection
            if new_zone.end < cur_zone.start or new_zone.start > cur_zone.end:
                i += 1
                continue

            # if current is contained in new
            if new_zone.start <= cur_zone.start and new_zone.end >= cur_zone.end:
                self.configuration.priority_zones.remove(cur_zone)
                continue

            # if new is contained in current
            if new_zone.start > cur_zone.start and new_zone.end < cur_zone.end:
                self.configuration.priority_zones.insert(
                    i + 1,
                    PriorityZone(new_zone.end + 1, cur_zone.end, cur_zone.priority),
                )
                cur_zone.end = new_zone.start - 1
                break

            # if new is intersecting from the left side
            if new_zone.end < cur_zone.end:
                cur_zone.start = new_zone.end + 1

            # if new is intersecting from the right side
            if new_zone.start > cur_zone.start:
                cur_zone.end = new_zone.start - 1

            i += 1

        # if new is contained in or intersecting from the left side into the last zone
        last_zone = self.configuration.priority_zones[-1]

        if new_zone.end >= last_zone.start:
            if new_zone.start >= last_zone.start:
                self.configuration.priority_zones.insert(
                    -1,
                    PriorityZone(last_zone.start, new_zone.start - 1, last_zone.priority),
                )
            last_zone.start = new_zone.end + 1

        i = 0

        while new_zone.start > self.configuration.priority_zones[i].start:
            i += 1

        self.configuration.priority_zones.insert(i, new_zone)

    def format_text(self, text: str, entry: PenlogEntry) -> FormattedText:
        """
        Returns a formatted text according to the properties of the corresponding penlog entry.

        :param text: The text to be formatted.
        :param entry: The penlog entry to which the text corresponds.
        :return: The formatted text.
        """
        if entry.tags is not None and "JSON" in entry.tags:
            return FormattedText(text, curses.color_pair(self.color_ids[PenlogPriority.ERROR]))

        text_format = curses.color_pair(self.color_ids[entry.priority])

        if entry.priority < PenlogPriority.INFO:
            text_format = text_format | curses.A_BOLD

        return FormattedText(text, text_format)

    def default_text(self, text: str) -> FormattedText:
        """
        Returns a formatted text with the content of the given text and the default formatting.

        :param text: The text.
        :return: The formatted text.
        """
        return FormattedText(text, curses.color_pair(0))

    def formatted_entry_simple(
        self, entry: PenlogEntry, entry_id: int, prefix: str
    ) -> list[DisplayEntry]:
        """
        Returns a list of formatted entries, each corresponding to a single line of the entry referred to by the
        given entry, as displayed in the console.
        This is a helper function to avoid code duplication.

        :param entry: The entry.
        :param entry_id: The id of the entry.
        :return: The formatted display entries for the given entry.
        """
        _, max_width = self.window.getmaxyx()
        residual_width = max_width - len(prefix) - 1

        user_defined_lines = entry.data.splitlines()

        if len(user_defined_lines) == 0:
            user_defined_lines = [""]

        terminal_width_defined_lines = []

        for line in user_defined_lines:
            if len(line) == 0:
                terminal_width_defined_lines.append("")

            for i in range(ceil(len(line) / residual_width)):
                terminal_width_defined_lines.append(
                    line[i * residual_width : (i + 1) * residual_width]
                )

        result = [
            DisplayEntry(
                [
                    self.default_text(prefix),
                    self.format_text(terminal_width_defined_lines[0], entry),
                ],
                entry_id,
                0,
            )
        ]

        for i, line in enumerate(terminal_width_defined_lines[1:]):
            result.append(
                DisplayEntry(
                    [
                        self.default_text(" " * len(prefix)),
                        self.format_text(line, entry),
                    ],
                    entry_id,
                    i + 1,
                )
            )

        return result

    def formatted_entry(self, entry_id: int) -> list[DisplayEntry]:
        """
        Returns a list of formatted entries, each corresponding to a single line of the entry referred to by the
        given entry id, as displayed in the console.

        :param entry_id: The id of the entry.
        :return: The formatted display entries for the given entry.
        """
        entry = self.entries[entry_id]
        _, max_width = self.window.getmaxyx()

        prefix = ""

        if self.use_prefix:
            if self.relative_timings:
                if self.reference_time <= entry.datetime:
                    delta = int((entry.datetime - self.reference_time).total_seconds() * 1000)
                else:
                    prefix += "-"
                    delta = int((self.reference_time - entry.datetime).total_seconds() * 1000)

                prefix += f"{delta // 86400000:01}d {delta % 86400000 // 3600000:02}:{delta % 3600000 // 60000:02}:{delta % 60000 // 1000:02}:{delta % 1000:03}"
                prefix = prefix.rjust(19)

            else:
                prefix += entry.datetime.strftime("%b %d %H:%M:%S.%f")[:-3]
            prefix += " "
            prefix += entry.module
            if entry.tags is not None:
                prefix += f" [{', '.join(entry.tags)}]"
            prefix += ": "

        result = self.formatted_entry_simple(entry, entry_id, prefix)

        if self.configuration.interpret:
            if entry.interpretation is None:
                self.interpret_entry(entry)

            if entry.interpretation is not None:
                interpretation_text = f"  # {entry.interpretation}"

                while len(interpretation_text) > 0:
                    len_texts = 0

                    for text in result[-1].texts:
                        len_texts += len(text.text)

                    residual_width = max_width - len_texts - 1

                    result[-1].texts.append(
                        FormattedText(
                            interpretation_text[:residual_width],
                            curses.color_pair(entry.interpretation_color),
                        )
                    )

                    interpretation_text = interpretation_text[residual_width:]

                    if len(interpretation_text) > 0:
                        result.append(
                            DisplayEntry(
                                [self.default_text(" " * len(prefix))],
                                entry_id,
                                result[-1].entry_line_number,
                            )
                        )

        result[-1].last_line = True

        return result

    def interpret_entry(self, entry: PenlogEntry) -> None:
        """
        Adds an interpretation to the entry, if the entry can be identified as referring to a UDS request or response.

        :param entry: The entry to be interpreted.
        """
        try:
            if not entry.data.startswith("00"):
                data = unhexlify(entry.data)

                if data[0] & 0b01000000:
                    response = UDSResponse.parse_dynamic(data)
                    entry.interpretation = repr(response)

                    if isinstance(response, NegativeResponse):
                        entry.interpretation_color = InterpretationColor.UDS_NEGATIVE_RESPONSE
                    else:
                        entry.interpretation_color = InterpretationColor.UDS_POSITIVE_RESPONSE
                else:
                    entry.interpretation = repr(UDSRequest.parse_dynamic(data))
                    entry.interpretation_color = InterpretationColor.UDS_REQUEST
        except Exception as e:
            self.debug_log(repr(e))

    def check_filter(self, entry_id: int) -> bool:
        """
        Check if the entry which is referred to by the given entry id, is still to be displayed after applying filters.

        :param entry_id: The index of the entry.
        :return: The formatted display entries for the given entry.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SyntaxWarning)

            for command in self.configuration.filter:
                if not eval(command, self.entries[entry_id].__dict__):
                    return False

        return True

    def calculate_display_entries(self, start_entry: int, entry_line: int) -> list[DisplayEntry]:
        """
        Returns a list of display entries starting from the given start entry, which qualify to be displayed by having
        a sufficient priority as well as passing any other filter.

        The display entries may start from a different entry than the one pointed to by the given parameter,
        in case that entry itself does not qualify itself.
        In that case previous entries are taken into account if the entry_line is set to -1
        and if there is at least oneprevious entry which qualifies.
        Otherwise only entries following the given one are taken into account.

        The number of entries which is returned is limited by the number of lines in the console.

        The list is guaranteed to be not empty.
        If there would be no entries to be displayed there will be an info message,
        which informs the user, that no entries are available with the current filtering settings.

        :param start_entry: The starting point for the search for the actual first entry (which qualifies) in the list.
        :param entry_line: The index in the list of formatted lines as created for the corresponding penlog entry.
                           Most importantly, a value of 0 indicates to start from the first line,
                           a value of -1 to start from the last line. Other negative values are not supported.
        :return: The list of display entries.
        """
        max_lines, _ = self.window.getmaxyx()
        max_lines -= 1
        display_entries: list[DisplayEntry] = []

        prio = self.entries[start_entry].priority
        zone = self.entry_zone(start_entry)

        fallback_entry = [
            DisplayEntry(
                [self.default_text("No entries found! You can select a new priority on this line")],
                0,
                0,
            )
        ]

        if prio > zone.priority or not self.check_filter(start_entry):
            if entry_line == -1:
                start_entry_ = self.previous_sufficient_entry(start_entry)

                if start_entry_ is None:
                    start_entry_ = self.next_sufficient_entry(0)

                    if start_entry_ is None:
                        return fallback_entry

                    entry_line = 0

                start_entry = start_entry_
            elif entry_line == 0:
                start_entry_ = self.next_sufficient_entry(start_entry)

                if start_entry_ is None:
                    return fallback_entry

                start_entry = start_entry_

            zone = self.entry_zone(start_entry)

        pointer = self.priority_pointer(start_entry, zone.priority)
        assert pointer is not None, f"{start_entry} {zone.priority}"

        i = 0

        while True:
            entry_id = self.level_pointers[zone.priority][pointer]
            new_lines = self.formatted_entry(entry_id)

            if i == 0:
                if entry_line == -1:
                    display_entries.append(new_lines[-1])
                else:
                    display_entries += new_lines[entry_line:]
            else:
                display_entries += new_lines

            if len(display_entries) > max_lines:
                break
            i += 1

            zone, pointer = self.next_sufficient_pointer(zone, pointer, entry_id)

            if pointer is None:
                break

        return display_entries[:max_lines]

    def next_sufficient_entry(self, entry_id: int) -> int | None:
        """
        Returns the index of the next entry after the given one,
        which has a sufficiently high priority to be displayed and also passes any filters.

        If no such entry exists None is returned.

        :param entry_id: The index of the entry in self.entries.
        :return: The index of the next sufficient entry after the given one or None, if no such entry exists.
        """
        prio = self.entries[entry_id].priority
        pointer = self.priority_pointer(entry_id, prio)
        assert pointer is not None
        zone, pointer = self.next_sufficient_pointer(None, pointer, entry_id)
        return self.level_pointers[zone.priority][pointer] if pointer is not None else None

    def previous_sufficient_entry(self, entry_id: int) -> int | None:
        """
        Returns the index of the previous entry after the given one,
        which has a sufficiently high priority to be displayed and also passes any filters.

        If no such entry exists None is returned.

        :param entry_id: The index of the entry in self.entries.
        :return: The index of the previous sufficient entry before the given one or None, if no such entry exists.
        """
        prio = self.entries[entry_id].priority
        pointer = self.priority_pointer(entry_id, prio)
        assert pointer is not None
        zone, pointer = self.previous_sufficient_pointer(None, pointer, entry_id)
        return self.level_pointers[zone.priority][pointer] if pointer is not None else None

    def next_sufficient_pointer(
        self, prio_zone: PriorityZone | None, pointer: int, entry_id: int
    ) -> tuple[PriorityZone, int | None]:
        """
        Returns the index of the next priority pointer after the given one,
        which has a sufficiently high priority to be displayed and also passes any filters,
        along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The next priority pointer's index along the corresponding priority zone.
        """
        zone, ptr = self.next_sufficient_pointer_without_filters(prio_zone, pointer, entry_id)

        while ptr is not None and not self.check_filter(self.level_pointers[zone.priority][ptr]):
            entry_id = self.level_pointers[zone.priority][ptr]
            zone, ptr = self.next_sufficient_pointer_without_filters(prio_zone, ptr, entry_id)

        return zone, ptr

    def previous_sufficient_pointer(
        self, prio_zone: PriorityZone | None, pointer: int, entry_id: int
    ) -> tuple[PriorityZone, int | None]:
        """
        Returns the index of the previous priority pointer before the given one,
        which has a sufficiently high priority to be displayed and also passes any filters,
        along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The previous priority pointer's index along the corresponding priority zone.
        """
        zone, ptr = self.previous_sufficient_pointer_without_filters(prio_zone, pointer, entry_id)

        while ptr is not None and not self.check_filter(self.level_pointers[zone.priority][ptr]):
            entry_id = self.level_pointers[zone.priority][ptr]
            zone, ptr = self.previous_sufficient_pointer_without_filters(prio_zone, ptr, entry_id)

        return zone, ptr

    def next_sufficient_pointer_without_filters(
        self, prio_zone: PriorityZone | None, pointer: int, entry_id: int
    ) -> tuple[PriorityZone, int | None]:
        """
        Returns the index of the next priority pointer after the given one,
        which has a sufficiently high priority to be displayed, along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The next priority pointer's index along the corresponding priority zone.
        """
        for zone in self.configuration.priority_zones:
            if zone.end is not None and entry_id >= zone.end:
                continue

            prio_entries = self.level_pointers[zone.priority]

            if len(prio_entries) == 0:
                continue

            next_pointer: int | None

            if zone == prio_zone and pointer + 1 < len(prio_entries):
                next_pointer = pointer + 1
            else:
                next_pointer = self.priority_pointer(
                    max(entry_id, zone.start - 1), zone.priority, 1
                )

            if next_pointer is not None and (
                zone.end is None or prio_entries[next_pointer] <= zone.end
            ):
                return zone, next_pointer

        return self.configuration.priority_zones[-1], None

    def previous_sufficient_pointer_without_filters(
        self, prio_zone: PriorityZone | None, pointer: int, entry_id: int
    ) -> tuple[PriorityZone, int | None]:
        """
        Returns the index of the previous priority pointer before the given one,
        which has a sufficiently high priority to be displayed, along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The previous priority pointer's index along the corresponding priority zone.
        """
        for zone in reversed(self.configuration.priority_zones):
            if entry_id <= zone.start:
                continue

            prio_entries = self.level_pointers[zone.priority]

            if len(prio_entries) == 0:
                continue

            prev_pointer: int | None

            if zone == prio_zone and pointer > 0:
                prev_pointer = pointer - 1
            else:
                prev_pointer = self.priority_pointer(
                    min(entry_id, zone.end + 1) if zone.end is not None else entry_id,
                    zone.priority,
                    -1,
                )

            if prev_pointer is not None and prio_entries[prev_pointer] >= zone.start:
                return zone, prev_pointer

        return self.configuration.priority_zones[0], None

    def priority_pointer(self, entry_id: int, prio: PenlogPriority, mode: int = 0) -> int | None:
        """
        Returns the index of the level pointer of the entry with the given id,
        or depending on the mode the previous or next entry, in the level pointers with the given priority.

        Depending on the mode parameter different pointer indices can be retrieved:
         - case 0: The pointer which points to the exact same entry which is given as the entry id
         - case 1: - 1: The pointer with the next bigger entry_id in the level pointers compared to the given entry id
         - case -1: The pointer with the next smaller entry_id in the level pointers compared to the given entry id

         In none of these cases it can be guaranteed that such a pointer exists in the level pointers for the given
         priority. If no pointer exists, None is returned.

        :param entry_id: The index of the entry in self.entries.
        :param prio: The priority level for which the index in the corresponding level pointers should be calculated.
        :param mode: Determines which pointer index should be returned (must be any of 0, 1 or -1).
        :return: The index of the level pointer or None if no such pointer could be identified.
        """
        prio_entries = self.level_pointers[prio]

        upper = len(prio_entries) - 1
        lower = 0

        while True:
            pointer = (upper + lower) // 2

            if upper == lower + 1 and mode == -1:
                pointer = upper

            prio_entry = prio_entries[pointer]

            if upper == lower:
                break

            # Find the exact match
            if mode == 0:
                if prio_entry > entry_id:
                    upper = max(lower, pointer - 1)
                elif prio_entry < entry_id:
                    lower = min(pointer + 1, upper)
                else:
                    break

            # Find the next bigger entry
            if mode == 1:
                if prio_entry > entry_id:
                    upper = max(lower, pointer)
                else:
                    lower = min(pointer + 1, upper)

            # Find the next smaller entry
            if mode == -1:
                if prio_entry >= entry_id:
                    upper = max(lower, pointer - 1)
                else:
                    lower = min(pointer, upper)

        if mode == 0 and prio_entry != entry_id:
            return None

        if mode == 1 and prio_entry <= entry_id:
            return None

        if mode == -1 and prio_entry >= entry_id:
            return None

        return pointer

    def handle_io(self) -> None:
        start_entry: int | None = None
        display_help = False
        entry_start_saved = 0
        line_start_saved = 0
        cursor_saved = (0, 0)
        start_entry_saved = None
        max_lines, max_columns = self.window.getmaxyx()
        max_lines -= 1
        filter_history = [self.configuration.filter]

        display_entries = self.calculate_display_entries(0, 0)
        self.display(display_entries, self.status(display_entries, (len(display_entries) - 1, 0)))
        prefix_length = 0
        cursor = (0, prefix_length)
        self.window.move(*cursor)

        def update_selected_zones(prio: PenlogPriority) -> None:
            nonlocal start_entry
            nonlocal display_entries
            nonlocal cursor
            nonlocal max_lines

            if start_entry is None:
                return

            stop_display_entry = display_entries[min(cursor[0], len(display_entries) - 1)]
            stop_entry = stop_display_entry.penlog_entry_number

            if start_entry == stop_entry:
                start_entry = self.previous_sufficient_entry(start_entry)
                stop_entry_ = self.next_sufficient_entry(stop_entry)

                if start_entry is None:
                    start_entry = 0

                if stop_entry_ is None:
                    stop_entry = len(self.entries) - 1
                else:
                    stop_entry = stop_entry_

            self.update_zones(
                PriorityZone(min(start_entry, stop_entry), max(start_entry, stop_entry), prio)
            )
            start_entry = None

        while (key := self.window.getkey()) != "q" or display_help:
            entry_start = display_entries[0].penlog_entry_number
            line_start = display_entries[0].entry_line_number
            cursor = curses.getsyx()
            max_lines, max_columns = self.window.getmaxyx()
            max_lines -= 1

            def page_up() -> None:
                nonlocal entry_start
                nonlocal line_start
                nonlocal display_entries

                for _ in range(max_lines - 1):
                    old_entry_start = entry_start
                    old_line_start = line_start

                    if display_entries[0].entry_line_number == 0:
                        if entry_start > 0:
                            entry_start = max(0, display_entries[0].penlog_entry_number - 1)
                            line_start = -1
                    else:
                        line_start = display_entries[0].entry_line_number - 1

                    if old_entry_start == entry_start and old_line_start == line_start:
                        break

                    display_entries = self.calculate_display_entries(entry_start, line_start)

            def line_up() -> None:
                nonlocal entry_start
                nonlocal line_start

                if display_entries[0].entry_line_number == 0:
                    if entry_start > 0:
                        entry_start = max(0, entry_start - 1)
                        line_start = -1
                else:
                    line_start -= 1

            match key:
                case "KEY_UP":
                    if cursor[0] > 0:
                        cursor = (cursor[0] - 1, cursor[1])
                    else:
                        line_up()
                case "KEY_DOWN":
                    if cursor[0] < max_lines - 1:
                        cursor = cursor[0] + 1, cursor[1]
                    elif display_entries[0].last_line:
                        entry_start = min(len(self.entries) - 1, entry_start + 1)
                        line_start = 0
                    else:
                        line_start += 1
                case "KEY_PPAGE":
                    if cursor[0] > 0:
                        cursor = (0, cursor[1])
                    else:
                        page_up()
                case "KEY_NPAGE":
                    if cursor[0] < len(display_entries) - 1:
                        cursor = (len(display_entries) - 1, cursor[1])
                    else:
                        entry_start = display_entries[-1].penlog_entry_number
                        line_start = display_entries[-1].entry_line_number
                case "g":
                    entry_start = 0
                    line_start = 0
                    cursor = (0, cursor[1])
                case "G":
                    entry_start = len(self.entries) - 1
                    line_start = -1

                    display_entries = self.calculate_display_entries(entry_start, line_start)
                    page_up()
                    cursor = (len(display_entries) - 1, cursor[1])
                case "KEY_LEFT":
                    if cursor[1] > 0:
                        self.window.move(cursor[0], cursor[1] - 1)
                        continue
                case "KEY_RIGHT":
                    if cursor[1] < max_columns - 1:
                        self.window.move(cursor[0], cursor[1] + 1)
                        continue
                # TODO: this is chr(curses.ascii.ESC); but that's no pattern.
                case "\x1b" | "q":
                    start_entry = None

                    if display_help:
                        display_help = False
                        entry_start = entry_start_saved
                        line_start = line_start_saved
                        cursor = cursor_saved
                        start_entry = start_entry_saved

            if not display_help:
                match key:
                    case "v":
                        if cursor[0] < len(display_entries):
                            start_entry = display_entries[cursor[0]].penlog_entry_number
                    case "p" | "P":
                        function_key = key

                        while (key := self.window.getkey()) != "q":
                            if key == chr(curses.ascii.ESC):
                                start_entry = None
                                break

                            try:
                                prio = self.priority_keys[key]
                            except KeyError:
                                continue

                            if function_key == "p":
                                update_selected_zones(prio)
                            else:
                                self.new_configuration()
                                self.configuration.priority_zones = [PriorityZone(0, None, prio)]

                            break
                    case "u":
                        self.configuration_index = max(0, self.configuration_index - 1)
                    case "r":
                        self.configuration_index = min(
                            len(self.configuration_history) - 1,
                            self.configuration_index + 1,
                        )
                    case "i":
                        self.new_configuration()
                        self.configuration.interpret = not self.configuration.interpret
                    case "f":
                        # fh is short for filter_history to reduce long unreadable lines
                        fh_tmp = ["; ".join(filter_commands) for filter_commands in filter_history]
                        fh_tmp.append("")
                        fh_index = len(fh_tmp) - 1

                        filter_cursor = len(fh_tmp[fh_index])
                        input_format = curses.color_pair(0)

                        self.display(
                            display_entries,
                            [FormattedText(fh_tmp[fh_index], input_format)],
                        )
                        self.window.move(max_lines, filter_cursor)

                        while (key := self.window.getkey()) != chr(curses.ascii.ESC):
                            match key:
                                case "\n":
                                    try:
                                        filter_tmp = parse_filter(fh_tmp[fh_index])
                                        self.debug_log(filter_tmp)
                                        filter_history.append(filter_tmp)
                                        self.new_configuration()
                                        self.configuration.filter = filter_tmp
                                        break
                                    except Exception:
                                        pass

                                case "KEY_BACKSPACE":
                                    if filter_cursor > 0:
                                        fh_tmp[fh_index] = (
                                            fh_tmp[fh_index][: filter_cursor - 1]
                                            + fh_tmp[fh_index][filter_cursor:]
                                        )
                                        filter_cursor -= 1
                                case "KEY_DC":
                                    if filter_cursor < len(fh_tmp[fh_index]):
                                        fh_tmp[fh_index] = (
                                            fh_tmp[fh_index][:filter_cursor]
                                            + fh_tmp[fh_index][filter_cursor + 1 :]
                                        )
                                case "KEY_LEFT":
                                    if filter_cursor > 0:
                                        filter_cursor -= 1
                                case "KEY_RIGHT":
                                    if filter_cursor < len(fh_tmp[fh_index]):
                                        filter_cursor += 1
                                case "KEY_UP":
                                    fh_index = max(0, fh_index - 1)
                                case "KEY_DOWN":
                                    fh_index = min(len(fh_tmp) - 1, fh_index + 1)
                                case _:
                                    fh_tmp[fh_index] = (
                                        fh_tmp[fh_index][:filter_cursor]
                                        + key
                                        + fh_tmp[fh_index][filter_cursor:]
                                    )
                                    filter_cursor += 1

                            try:
                                parse_filter(fh_tmp[fh_index])
                                input_format = curses.color_pair(0)
                            except Exception:
                                input_format = curses.color_pair(
                                    self.color_ids[PenlogPriority.WARNING]
                                )

                            self.display(
                                display_entries,
                                [FormattedText(fh_tmp[fh_index], input_format)],
                            )
                            self.window.move(max_lines, min(filter_cursor, len(fh_tmp[fh_index])))
                    case "x":
                        self.use_prefix = not self.use_prefix
                    case "t":
                        self.relative_timings = not self.relative_timings
                    case "z":
                        self.reference_time = self.entries[
                            display_entries[cursor[0]].penlog_entry_number
                        ].datetime
                    case "?":
                        display_help = True
                        entry_start_saved = entry_start
                        line_start_saved = line_start
                        cursor_saved = cursor
                        start_entry_saved = start_entry
                        entry_start = 0
                        line_start = 0
                        cursor = (0, 0)
                        start_entry = None

            if display_help:
                display_entries = self.help_message(line_start)
            else:
                display_entries = self.calculate_display_entries(entry_start, line_start)

            previous_entry_start = 0
            previous_line_start = 0

            while len(display_entries) < max_lines and (
                previous_entry_start != entry_start or previous_line_start != line_start
            ):
                previous_entry_start = entry_start
                previous_line_start = line_start
                entry_start = display_entries[0].penlog_entry_number
                line_start = display_entries[0].entry_line_number
                line_up()

                if display_help:
                    display_entries = self.help_message(line_start)
                else:
                    display_entries = self.calculate_display_entries(entry_start, line_start)

            if start_entry is not None:
                stop_entry = display_entries[
                    min(cursor[0], len(display_entries) - 1)
                ].penlog_entry_number

                for entry in display_entries:
                    if (
                        min(start_entry, stop_entry)
                        <= entry.penlog_entry_number
                        <= max(start_entry, stop_entry)
                    ):
                        for text in entry.texts:
                            text.format = text.format | curses.A_REVERSE

            status = [] if display_help else self.status(display_entries, cursor)
            self.display(display_entries, status)
            self.window.move(min(cursor[0], len(display_entries) - 1), cursor[1])

    def help_message(self, line_start: int) -> list[DisplayEntry]:
        options = {
            "f": "Enter filter input mode (beware!, this is executed with eval on each line)",
            "g": "Jump to the start of the file",
            "G": "Jump to the end of the file",
            "i": "Interpret UDS messages (they appear as comments next to the original message)",
            "p": "Change the log level (priority) for the selected range, followed by the corresponding log level key",
            "P": "Change the log level (priority) for the entire file, followed by the corresponding log level key",
            "q": "Quit the application or this very help message",
            "r": "Redo the last undone action (as long as no new action has been done)",
            "t": "Toggle between absolute timestamps and timings relative to the reference timestamp (default: start)",
            "u": "Undo the last action",
            "v": "Start marking ranges (for further actions)",
            "x": "Toggle prefix displaying",
            "z": "Set the reference timestamp to the timestamp of the entry under the cursor",
            "?": "Show this help message",
        }

        display_entries: list[DisplayEntry] = []

        def add_entries(message: str, prefix: str = "") -> None:
            nonlocal display_entries

            entry = PenlogEntry(
                data=message,
                datetime=datetime.now(),
                host="",
                module="",
                priority=PenlogPriority.INFO,
            )
            display_entries += self.formatted_entry_simple(entry, 0, "    " + prefix)

        add_entries("")
        add_entries("HELP")
        add_entries("")
        add_entries("Move cursor with arrow keys or Page Up / Page Down")
        add_entries("Press ESC to cancel an action or go back to the main view")
        add_entries("")
        add_entries("Action keys:")

        for key, val in options.items():
            add_entries(f"{val}", f"    {key}: ")

        add_entries("")
        add_entries("Log level keys (ordered from highest to lowest):")

        for level in self.priority_keys:
            add_entries(f"{PenlogPriority(self.priority_keys[level]).name}", f"    {level}: ")

        add_entries("")
        add_entries("Filtering information:")
        add_entries(
            "Beware that filters are arbitrary Python statements executed with eval on each line!"
        )
        add_entries("This can lead to poor performance on large files as well as side effects!")
        add_entries("Existing filters can be traversed using the up and down arrow keys")
        add_entries("The following attributes of each line are exposed as a variable:")

        for attr in self.entries[0].__dict__:
            if not attr.startswith("_"):
                add_entries(f"{attr}", "    ")

        max_lines, _ = self.window.getmaxyx()

        for n, entry in enumerate(display_entries):
            entry.entry_line_number = n

        return display_entries[line_start : max_lines + line_start]

    def status(
        self, display_entries: list[DisplayEntry], cursor: tuple[int, int]
    ) -> list[FormattedText]:
        selected_entry = display_entries[min(cursor[0], len(display_entries) - 1)]
        entry_number = selected_entry.penlog_entry_number + 1
        n = len(self.entries)
        zone = self.entry_zone(selected_entry.penlog_entry_number)
        zone_info = self.default_text(
            f"Zone {self.configuration.priority_zones.index(zone): >2} ({zone.priority.name})"
        )
        progress = self.default_text(
            f"{entry_number: >{len(str(n))}}/{n} ({(entry_number / n) * 100:6.2f}%)"
        )
        _, max_columns = self.window.getmaxyx()
        spacing = self.default_text(
            " " * (max_columns - 1 - len(zone_info.text) - len(progress.text))
        )
        return [zone_info, spacing, progress]

    def display(
        self, display_entries: list[DisplayEntry], status_line: list[FormattedText]
    ) -> None:
        self.window.erase()

        for i, display_entry in enumerate(display_entries):
            for text in display_entry.texts:
                try:
                    self.window.addstr(text.text, text.format)
                except ValueError:
                    self.window.addstr(text.sanitized_text, text.format)

            if i < len(display_entries) - 1:
                self.window.addstr("\n")

        self.window.move(self.window.getmaxyx()[0] - 1, 0)

        for text in status_line:
            try:
                self.window.addstr(text.text, text.format)
            except ValueError:
                self.window.addstr(text.sanitized_text, text.format)

    def debug_log(self, msg: Any) -> None:
        debug_path = Path("/tmp/cursed_log")

        try:
            with debug_path.open("a", encoding="utf-8") as f:
                f.write(str(msg) + "\n")
        except Exception:
            pass


def parse_filter(text: str) -> list[str]:
    test_entry = PenlogEntry(
        module="component",
        data="data",
        host="host",
        priority=PenlogPriority.INFO,
        datetime=datetime.fromtimestamp(0),
        tags=[],
    )

    commands = [command.strip() for command in text.split(";") if len(command.strip()) > 0]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", SyntaxWarning)

        for command in commands:
            eval(command, test_entry.__dict__)

    return commands


def main() -> None:
    parser = ArgumentParser()
    parser.add_argument("file", type=Path)
    parser.add_argument(
        "--priority", "-p", type=PenlogPriority.from_str, default=PenlogPriority.NOTICE
    )
    parser.add_argument("--filter", "-f", type=parse_filter, default=None)
    parser.add_argument("--prefix", action=BooleanOptionalAction, default=True)
    parser.add_argument("--relative-timings", action=BooleanOptionalAction, default=False)
    args = parser.parse_args()
    CursedHR(args.file, args.priority, args.filter, args.prefix, args.relative_timings)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Pydantic Utility Functions for Declarative Typed Argument Parsing.

The `pydantic` module contains utility functions used for interacting with the
internals of `pydantic`, such as constructing field validators, updating
field validator dictionaries and constructing new model classes with
dynamically generated validators and environment variable parsers.
"""

from collections.abc import Container, Iterator
from dataclasses import dataclass
from enum import Enum
from types import UnionType
from typing import (
    Annotated,
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from gallia.pydantic_argparse.utils.field import ArgFieldInfo

from .types import all_types

# Constants
T = TypeVar("T")
PydanticModelT = TypeVar("PydanticModelT", bound=BaseModel)
PydanticValidator = classmethod
NoneType = type(None)


@dataclass
class PydanticField:
    """Simple Pydantic v2.0 field wrapper.

    Pydantic fields no longer store their name, so this named tuple
    keeps the field name and field info together.

    The recommended entry point for an arbitrary `pydantic.BaseModel` is the classmethod `PydanticField.parse_model`.
    """

    name: str
    info: FieldInfo
    extra_default: tuple[str, Any] | None = None

    @classmethod
    def parse_model(cls, model: BaseModel | type[BaseModel]) -> Iterator["PydanticField"]:
        """Iterator over the pydantic model fields, yielding this wrapper class.

        Yields:
            Instances of self (`PydanticField`)
        """
        for name, info in model.model_fields.items():
            yield cls(name, info)

    def _get_type(self, annotation: type | None) -> type | tuple[type | None, ...] | None:
        origin = get_origin(annotation)

        if origin is Literal or isinstance(origin, type) and issubclass(origin, Container):
            return origin
        elif origin is Union or origin is UnionType:
            args = get_args(annotation)
            types = [arg for arg in args if arg is not NoneType]
        elif origin is None:
            types = [annotation]
        else:
            raise AssertionError(
                f"Unsupported origin {origin} for field {self.name} with annotation {annotation}"
            )

        base_types: list[type | None] = []

        for t in types:
            origin = get_origin(t)

            if origin is Annotated:
                sub_types = self._get_type(get_args(t)[0])

                if isinstance(sub_types, tuple):
                    base_types += sub_types
                else:
                    base_types.append(sub_types)
            elif origin is not None:
                base_types.append(origin)
            else:
                base_types.append(t)

        return tuple(base_types)

    def get_type(self) -> type | tuple[type | None, ...] | None:
        """Return the mainly interesting types according to the type annotation (in the context of argument parsing).

        Returns: One or more types (potentially None).
        """
        annotation = self.info.annotation
        return self._get_type(annotation)

    def is_a(self, types: Any | tuple[Any, ...]) -> bool:
        """Checks whether the subject *is* any of the supplied types.

        The checks are performed as follows:

        1. `field` *is* one of the `types`
        2. `field` *is an instance* of one of the `types`
        3. `field` *is a subclass* of one of the `types`

        If any of these conditions are `True`, then the function returns `True`,
        else `False`.

        Args:
            types (Union[Any, Tuple[Any, ...]]): Type(s) to compare field against.

        Returns:
            bool: Whether the field *is* considered one of the types.
        """
        # Create tuple if only one type was provided
        if not isinstance(types, tuple):
            types = (types,)

        # Get field type, or origin if applicable
        field_type = self.get_type()
        if not isinstance(field_type, tuple):
            field_type = (field_type,)

        # Check `isinstance` and `issubclass` validity
        # In order for `isinstance` and `issubclass` to be valid, all arguments
        # should be instances of `type`, otherwise `TypeError` *may* be raised.

        is_valid = all_types((*types, *field_type))

        # Perform checks and return
        is_type = False
        for t in field_type:
            is_type = (
                is_type
                or t in types
                or (is_valid and isinstance(t, types))
                or (is_valid and issubclass(t, types))  # type: ignore
            )

        return is_type

    @property
    def model_type(self) -> type[BaseModel]:
        """Try to return the `pydantic.BaseModel` type.

        Raises:
            TypeError: if this field is not a `pydantic.BaseModel` or if the model type cannot be found.
        """
        if not self.is_a(BaseModel):
            raise TypeError("This `pydantic` field is not a `pydantic.BaseModel`")

        types = self.get_type()
        if not isinstance(types, tuple):
            return cast(type[BaseModel], types)

        for t in types:
            if isinstance(t, type) and issubclass(t, BaseModel):
                return t
        raise TypeError("No `pydantic.BaseModel`s were found associated with this field.")

    def is_subcommand(self) -> bool:
        """Check whether the input pydantic Model is a subcommand.

        The default is that all pydantic Models are not subcommands, so this
        has this featured has to be opted in by adding `subcommand=True`
        to the `json_schema_extra` model config. A convenience class has been created
        to provide this default for models that are meant to be command switches at
        the command line: `pydantic_argparse.BaseCommand`.

        Returns:
            bool: if the pydantic model is a subcommand. In all other cases, including when this field is not a
                pydantic model, returns False.
        """
        default = False
        try:
            model = self.model_type
            value = model.model_config["json_schema_extra"].get("subcommand", default)  # type: ignore
            return cast(bool, value)
        except (KeyError, AttributeError, TypeError):
            # KeyError if:
            #   - subcommand key not in json_schema_extra

            # AttributeError if:
            #   - json_schema_extra not in the model_config, ie if using BaseModel
            # just default to not being a subcommand

            # TypeError if
            #   - field is not a pydantic BaseModel or it can't be found
            return default

    def arg_names(self, invert: bool = False) -> tuple[str, str] | tuple[str]:
        """Standardises argument name when printing to command line.

        This also includes potential short names if specified.

        Args:
            invert (bool): Whether to invert the name by prepending `--no-`.

        Returns:
            str: Standardised name of the argument.
        """
        name = self.info.title or self.name

        if isinstance(self.info, ArgFieldInfo) and self.info.positional:
            return (name,)

        prefix = "--no-" if invert else "--"
        long_name = f"{prefix}{name.replace('_', '-')}"

        if isinstance(self.info, ArgFieldInfo) and self.info.short is not None:
            return f"-{self.info.short}", long_name

        return (long_name,)

    def description(self) -> str:
        """Standardises argument description.

        Returns:
            str: Standardised description of the argument.
        """
        # Construct Default String
        default = ""

        if not self.info.is_required():
            _default = self.info.get_default()
            if isinstance(_default, Enum):
                _default = _default.name
            default = f"default: {_default}"

        if self.extra_default is not None:
            if len(default) > 0:
                default += "; "
            default += f"{self.extra_default[0]}: {self.extra_default[1]}"

        if len(default) > 0:
            default = f" ({default})"

        # Return Standardised Description String
        description = self.info.description if self.info.description is not None else ""
        return f"{description}{default}"

    def metavar(self) -> str | None:
        """Generate the metavar name for the field.

        Returns:
            Optional[str]: Field metavar if of type ArgField and has metavar set.
                Otherwise, return constituent type names.
        """
        # check metavar first
        if isinstance(self.info, ArgFieldInfo):
            if self.info.metavar is not None:
                return self.info.metavar

            if self.info.positional:
                return self.arg_names()[0].upper()

        # otherwise default to the type
        field_type = self.get_type()
        if field_type is not None:
            if isinstance(field_type, tuple):
                return "|".join(t.__name__.upper() for t in field_type if t is not None)
            return field_type.__name__.upper()
        else:
            return None

    def arg_required(self) -> dict[str, bool]:
        return (
            {}
            if isinstance(self.info, ArgFieldInfo) and self.info.positional
            else {"required": self.info.is_required() and self.extra_default is None}
        )

    def arg_default(self) -> dict[str, Any]:
        return (
            {}
            if self.extra_default is None
            or isinstance(self.info, ArgFieldInfo)
            and self.info.positional
            else {"default": self.extra_default[1]}
        )

    def arg_const(self) -> dict[str, Any]:
        return (
            {"const": self.info.const, "nargs": "?"}
            if isinstance(self.info, ArgFieldInfo) and self.info.const is not PydanticUndefined
            else {}
        )

    def arg_dest(self) -> dict[str, str]:
        return (
            {}
            if isinstance(self.info, ArgFieldInfo) and self.info.positional
            else {"dest": self.name}
        )


def is_subcommand(model: BaseModel | type[BaseModel]) -> bool:
    """Check whether the input pydantic Model is a subcommand.

    The default is that all pydantic Models are not subcommands, so this
    has this featured has to be opted in by adding `subcommand=True`
    to the `json_schema_extra` model config. A convenience class has been created
    to provide this default for models that are meant to be command switches at
    the command line: `pydantic_argparse.BaseCommand`.


    Args:
        model (BaseModel | Type[BaseModel]): a pydantic BaseModel subclass

    Returns:
        bool: if the pydantic model is a subcommand
    """
    default = False
    try:
        value = model.model_config["json_schema_extra"].get("subcommand", default)  # type: ignore
        return cast(bool, value)
    except (KeyError, AttributeError):
        # KeyError if:
        #   - subcommand key not in json_schema_extra
        # AttributeError if:
        #   - json_schema_extra not in the model_config, ie if using BaseModel
        # just default to not being a subcommand
        return default



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import base64
import pickle
import sys

from gallia.command.base import AsyncScriptConfig
from gallia.command.config import AutoInt, Field, Ranges

assert sys.platform == "win32"

from gallia.command import AsyncScript
from gallia.transports._ctypes_vector_xl_wrapper import FlexRayCtypesBackend
from gallia.transports.flexray_vector import FlexRayFrame, RawFlexRayTransport, parse_frame_type


class FRDumpConfig(AsyncScriptConfig):
    target_slot: AutoInt | None = Field(description="the target flexray slot")
    isotp: bool = Field(False, description="the target flexray slot")
    filter_null_frames: bool = Field(True, description="filter mysterious null frames out")
    slot: Ranges = Field([], description="filter on flexray slot")


class FRDump(AsyncScript):
    """Dump the content of the flexray bus"""

    CONFIG_TYPE = FRDumpConfig
    SHORT_HELP = "runs a helper tool that dumps flexray bus traffic to stdout"

    def __init__(self, config: FRDumpConfig):
        super().__init__(config)
        self.config: FRDumpConfig = config

    @staticmethod
    def poor_mans_dissect(frame: FlexRayFrame) -> str:
        res = f"slot_id: {frame.slot_id:03d}; "

        hdr = frame.data[:4]
        res += f"hdr: {hdr.hex()}; "

        frame_type = parse_frame_type(frame.data[4:])
        res += f"type: {frame_type.name}; "
        res += f"data: {frame.data[4:].hex()}"

        return res

    async def main(self) -> None:
        tp = await RawFlexRayTransport.connect("fr-raw:", None)

        if len(self.config.slot) > 0:
            tp.add_block_all_filter()
            for slot in self.config.slot:
                tp.set_acceptance_filter(slot)

        tp.activate_channel()

        while True:
            frame = await tp.read_frame()

            if self.config.filter_null_frames is True:
                # Best effort; in our use case this was the ISO-TP header.
                # The first ISO-TP header byte is never 0x00.
                if frame.data[0] == 0x00:
                    continue

            if self.config.isotp:
                print(self.poor_mans_dissect(frame))
            else:
                print(f"slot_id: {frame.slot_id:03d}; data: {frame.data.hex()}")


class FRConfigDumpConfig(AsyncScriptConfig):
    channel: int | None = Field(description="the channel number of the flexray device")
    pretty: bool = Field(False, description="pretty print the configuration", short="p")


class FRConfigDump(AsyncScript):
    """Dump the flexray configuration as base64"""

    CONFIG_TYPE = FRConfigDumpConfig
    SHORT_HELP = "Dump the flexray configuration as base64"

    def __init__(self, config: FRConfigDumpConfig):
        super().__init__(config)
        self.config: FRConfigDumpConfig = config

    async def main(self) -> None:
        backend = FlexRayCtypesBackend.create(self.config.channel)
        raw_config = backend.get_configuration()
        config = pickle.dumps(raw_config)

        if self.config.pretty:
            print(raw_config)
        else:
            print(base64.b64encode(config))



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import reprlib
import sys
from typing import Any

from gallia.command import UDSScanner
from gallia.command.config import AutoInt, Field, Ranges, Ranges2D
from gallia.command.uds import UDSScannerConfig
from gallia.log import get_logger
from gallia.services.uds import (
    NegativeResponse,
    UDSErrorCodes,
    UDSIsoServices,
    UDSRequestConfig,
    UDSResponse,
)
from gallia.services.uds.core.exception import (
    MalformedResponse,
    RequestResponseMismatch,
    UDSException,
)
from gallia.services.uds.core.utils import g_repr

logger = get_logger(__name__)


class ServicesScannerConfig(UDSScannerConfig):
    sessions: Ranges | None = Field(
        None,
        description="Set list of sessions to scan in; if None, services are scanned without session changes",
    )
    check_session: bool = Field(
        False, description="check current session; only takes effect if --sessions is given"
    )
    scan_response_ids: bool = Field(False, description="Include IDs in scan with reply flag set")
    reset: AutoInt | None = Field(
        None,
        description="Reset the ECU after every session with the optionally given reset level; only takes effect if --sessions is given",
        const=0x01,
    )
    skip: Ranges2D = Field(
        {},
        metavar="SESSION_ID:ID",
        description="\nThe service IDs to be skipped per session.\nA session specific skip is given by <session id>:<service ids>\nwhere <service ids> is a comma separated list of single ids or id ranges using a dash.\nExamples:\n - 0x01:0xf3\n - 0x10-0x2f\n - 0x01:0xf3,0x10-0x2f\nMultiple session specific skips are separated by space.\nOnly takes affect if --sessions is given.\n",
    )


class ServicesScanner(UDSScanner):
    """Iterate sessions and services and find endpoints"""

    CONFIG_TYPE = ServicesScannerConfig
    SHORT_HELP = "service scan on an ECU"
    EPILOG = "https://fraunhofer-aisec.github.io/gallia/uds/scan_modes.html#service-scan"

    def __init__(self, config: ServicesScannerConfig):
        super().__init__(config)
        self.config: ServicesScannerConfig = config

        self.result: list[tuple[int, int]] = []

    async def main(self) -> None:
        self.ecu.max_retry = 0
        found: dict[int, dict[int, Any]] = {}
        clean_returns = True

        if self.config.sessions is None:
            # Attempt to read session once for state update
            try:
                current_session = await self.ecu.read_session(
                    UDSRequestConfig(
                        timeout=2,
                        max_retry=0,
                    )
                )
                logger.notice(f"Performing scan in current session, which is {current_session:#x}")
            except Exception as e:
                current_session = 0
                logger.notice(f"Performing scan in current (unknown) session: {e!r}")

            found[current_session], clean_returns = await self.perform_scan()

        else:
            sessions = [
                s
                for s in self.config.sessions
                if s not in self.config.skip or self.config.skip[s] is not None
            ]
            logger.info(f"testing sessions {g_repr(sessions)}")

            # TODO: Unified shortened output necessary here
            logger.info(f"skipping identifiers {reprlib.repr(self.config.skip)}")

            for session in sessions:
                logger.info(f"Changing to session {g_repr(session)}")
                try:
                    resp: UDSResponse = await self.ecu.set_session(
                        session, UDSRequestConfig(tags=["preparation"])
                    )
                except (UDSException, RuntimeError) as e:  # FIXME: why catch RuntimeError?
                    logger.warning(
                        f"Could not complete session change to {g_repr(session)}: {g_repr(e)}; skipping session"
                    )
                    clean_returns = False
                    continue
                if isinstance(resp, NegativeResponse):
                    logger.warning(
                        f"Could not complete session change to {g_repr(session)}: {resp}; skipping session"
                    )
                    clean_returns = False
                    continue

                logger.result(f"scanning in session {g_repr(session)}")

                found[session], ret = await self.perform_scan(session)
                clean_returns = clean_returns and ret

                if self.config.reset is not None:
                    try:
                        logger.info("Resetting the ECU as requested")
                        reset_resp = await self.ecu.ecu_reset(self.config.reset)

                        if isinstance(reset_resp, NegativeResponse):
                            logger.warning(
                                f"Could not reset ECU: {reset_resp}; continuing without reset"
                            )
                        else:
                            logger.info("Waiting for the ECU to recover…")
                            await self.ecu.wait_for_ecu()
                    except (TimeoutError, ConnectionError):
                        logger.warning(
                            "Lost connection to the ECU after performing a reset. Attempting to reconnect…"
                        )
                        await self.ecu.reconnect()

        for key, value in found.items():
            logger.result(f"findings in session 0x{key:02X}:")
            for sid, data in value.items():
                self.result.append((key, sid))
                try:
                    logger.result(f"  [{g_repr(sid)}] {UDSIsoServices(sid).name}: {data}")
                except Exception:
                    logger.result(f"  [{g_repr(sid)}] vendor specific sid: {data}")

        if not clean_returns:
            sys.exit(1)

    async def perform_scan(self, session: None | int = None) -> tuple[dict[int, Any], bool]:
        result: dict[int, Any] = {}
        clean_returns = True

        # Starts at 0x00, see first loop iteration.
        sid = -1
        while sid < 0xFF:
            sid += 1
            if sid & 0x40 and (not self.config.scan_response_ids):
                continue

            if session in self.config.skip and (
                (session_skip := self.config.skip[session]) is None or sid in session_skip
            ):
                logger.info(f"{g_repr(sid)}: skipped")
                continue

            if session is not None and self.config.check_session:
                if not await self.ecu.check_and_set_session(session):
                    logger.error(
                        f"Aborting scan on session {g_repr(session)}; current SID was {g_repr(sid)}"
                    )
                    return result, False

            for length_payload in [1, 2, 3, 5]:
                pdu = bytes([sid]) + bytes(length_payload)
                try:
                    resp = await self.ecu.send_raw(pdu, config=UDSRequestConfig(tags=["ANALYZE"]))
                except TimeoutError:
                    logger.info(f"{g_repr(sid)}: timeout")
                    continue
                except (MalformedResponse, RequestResponseMismatch) as e:
                    logger.warning(f"{g_repr(sid)}: {e!r} occurred, this needs to be investigated!")
                    clean_returns = False
                    continue

                if isinstance(resp, NegativeResponse) and resp.response_code in [
                    UDSErrorCodes.serviceNotSupported,
                    UDSErrorCodes.serviceNotSupportedInActiveSession,
                ]:
                    logger.info(f"{g_repr(sid)}: not supported [{resp}]")
                    break

                if isinstance(resp, NegativeResponse) and resp.response_code in [
                    UDSErrorCodes.incorrectMessageLengthOrInvalidFormat
                ]:
                    continue

                logger.result(f"{g_repr(sid)}: available in session {g_repr(session)}: {resp}")
                result[sid] = resp
                break

        return result, clean_returns



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import sys

from gallia.transports.base import BaseTransport, TargetURI
from gallia.transports.doip import DoIPTransport
from gallia.transports.dummy import DummyTransport
from gallia.transports.hsfz import HSFZTransport
from gallia.transports.schemes import TransportScheme
from gallia.transports.tcp import TCPLinesTransport, TCPTransport

registry: list[type[BaseTransport]] = [
    DoIPTransport,
    DummyTransport,
    HSFZTransport,
    TCPLinesTransport,
    TCPTransport,
]

__all__ = [
    "BaseTransport",
    "DoIPTransport",
    "DummyTransport",
    "HSFZTransport",
    "TCPLinesTransport",
    "TCPTransport",
    "TargetURI",
    "TransportScheme",
]

if sys.platform.startswith("linux"):
    from gallia.transports.can import RawCANTransport
    from gallia.transports.isotp import ISOTPTransport
    from gallia.transports.unix import UnixLinesTransport, UnixTransport

    registry.append(ISOTPTransport)
    registry.append(RawCANTransport)
    registry.append(UnixLinesTransport)
    registry.append(UnixTransport)

    __all__ += [
        "ISOTPTransport",
        "RawCANTransport",
        "UnixLinesTransport",
        "UnixTransport",
    ]


if sys.platform == "win32":
    from gallia.transports.flexray_vector import FlexRayTPLegacyTransport, RawFlexRayTransport

    registry.append(RawFlexRayTransport)
    registry.append(FlexRayTPLegacyTransport)

    __all__ += ["FlexRayTPTransport", "FlexRayTPLegacyTransport"]



# This file has been autogenerated by `just gen-constants`.
# !! DO NOT CHANGE MANUALLY !!

# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import struct

CANFD_MTU = 72
CAN_MTU = 16

# https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/include/uapi/linux/can.h
CAN_HEADER_FMT = struct.Struct("=IBB2x")

CANFD_BRS = 0x01
CANFD_ESI = 0x02

CAN_EFF_FLAG = 2147483648
CAN_ERR_FLAG = 536870912
CAN_RTR_FLAG = 1073741824

CAN_EFF_MASK = 536870911
CAN_INV_FILTER = 0x20000000  # TODO: Add to CPython
CAN_SFF_MASK = 2047

CAN_RAW = 1
CAN_RAW_FD_FRAMES = 5
CAN_RAW_FILTER = 1
CAN_RAW_JOIN_FILTERS = 6
SOL_CAN_RAW = 101

# Socket Constants not available in the socket module,
# see linux/can/isotp.h
# TODO: This is now available in Python 3.15+.
# https://github.com/python/cpython/pull/23794
SOL_CAN_ISOTP = 106

# Valuetypes for SOL_CAN_ISOTP
CAN_ISOTP_OPTS = 1
CAN_ISOTP_RECV_FC = 2
CAN_ISOTP_TX_STMIN = 3
CAN_ISOTP_RX_STMIN = 4
CAN_ISOTP_LL_OPTS = 5

# Flags for setsockopt CAN_ISOTP_OPTS
CAN_ISOTP_LISTEN_MODE = 0x001
CAN_ISOTP_EXTEND_ADDR = 0x002
CAN_ISOTP_TX_PADDING = 0x004
CAN_ISOTP_RX_PADDING = 0x008
CAN_ISOTP_CHK_PAD_LEN = 0x010
CAN_ISOTP_CHK_PAD_DATA = 0x020
CAN_ISOTP_HALF_DUPLEX = 0x040
CAN_ISOTP_FORCE_TXSTMIN = 0x080
CAN_ISOTP_FORCE_RXSTMIN = 0x100
CAN_ISOTP_RX_EXT_ADDR = 0x200



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Utilities for Declarative Typed Argument Parsing.

This package contains helper utility functions for the typed argument parsing
process, including formatting argument names and descriptions, formatting
errors, recursively parsing `argparse.Namespace` objects to `dict`s,
interacting with the internals of `pydantic` and determining the types of
`pydantic` fields.

The public interface exposed by this package is the various described utility
modules each containing helper functions.
"""



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import asyncio
from asyncio import Task
from dataclasses import asdict, dataclass
from datetime import UTC, datetime
from enum import Enum
from json import JSONEncoder, dumps
from typing import TYPE_CHECKING, Any

from gallia.db.log import LogMode
from gallia.log import get_logger
from gallia.power_supply import PowerSupply
from gallia.services.uds.core import service
from gallia.services.uds.core.client import UDSClient, UDSRequestConfig
from gallia.services.uds.core.constants import DataIdentifier
from gallia.services.uds.core.exception import (
    ResponseException,
    UDSException,
    UnexpectedNegativeResponse,
)
from gallia.services.uds.core.utils import from_bytes, g_repr
from gallia.services.uds.helpers import (
    as_exception,
    raise_for_error,
    suggests_identifier_not_supported,
)
from gallia.transports.base import BaseTransport
from gallia.utils import handle_task_error, set_task_handler_ctx_variable

if TYPE_CHECKING:
    from gallia.db.handler import DBHandler


class ECUState:
    def __init__(self) -> None:
        self.session: int | None = None
        self.security_access_level: int | None = None

    def reset(self) -> None:
        self.session = None
        self.security_access_level = None

    def to_dict(self) -> dict[str, Any]:
        return self.__dict__

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(f'{key}={g_repr(value)}' for key, value in self.__dict__.items())})"


logger = get_logger(__name__)


@dataclass
class ECUProperties:
    def to_json(self, indent: int | str | None = None) -> str:
        # Make sure to keep 'sort_keys=True' when overriding this method!
        return dumps(asdict(self), indent=indent, sort_keys=True, cls=self.json_encoder)

    @property
    def json_encoder(self) -> type[JSONEncoder]:
        return ECUPropertiesEncoder


class ECUPropertiesEncoder(JSONEncoder):
    def default(self, obj):  # type: ignore
        if isinstance(obj, bytes):
            return obj.hex()
        elif isinstance(obj, Enum):
            return obj.value

        return super().default(obj)


class ECU(UDSClient):
    """ECU is a high level interface wrapping a UDSClient class. It provides
    semantically correct higher level interfaces such as read_session()
    or ping(). Vendor specific implementations can be derived from this
    class. For the arguments of the constructor, please check uds.uds.UDS.
    """

    OEM = "default"

    def __init__(
        self,
        transport: BaseTransport,
        timeout: float,
        max_retry: int = 0,
        power_supply: PowerSupply | None = None,
    ) -> None:
        super().__init__(transport, timeout, max_retry)
        self.tester_present_task: Task[None] | None = None
        self.tester_present_interval: float | None = None
        self.power_supply = power_supply
        self.state = ECUState()
        self.db_handler: DBHandler | None = None
        self.implicit_logging = True

    async def properties(
        self, fresh: bool = False, config: UDSRequestConfig | None = None
    ) -> ECUProperties:
        return ECUProperties()

    async def ping(
        self, config: UDSRequestConfig | None = None
    ) -> service.NegativeResponse | service.TesterPresentResponse:
        """Send an UDS TesterPresent message.

        Returns:
            UDS response.
        """
        return await self.tester_present(suppress_response=False, config=config)

    async def read_session(self, config: UDSRequestConfig | None = None) -> int:
        """Read out current session.

        Returns:
            The current session as int.
        """
        resp = await self.read_data_by_identifier(
            DataIdentifier.ActiveDiagnosticSessionDataIdentifier, config=config
        )
        if isinstance(resp, service.NegativeResponse):
            raise as_exception(resp)
        return from_bytes(resp.data_record)

    async def set_session_pre(self, level: int, config: UDSRequestConfig | None = None) -> bool:
        """set_session_pre() is called before the diagnostic session control
        pdu is written on the wire. Implement this if there are special
        preconditions for a particular session, such as disabling error
        logging.

        Args:
            uds: The UDSClient class where this hook is embedded. The caller typically
                 calls this function with `self` as the first argument.
            session: The desired session identifier.
        Returns:
            True on success, False on error.
        """
        return True

    async def set_session_post(self, level: int, config: UDSRequestConfig | None = None) -> bool:
        """set_session_post() is called after the diagnostic session control
        pdu was written on the wire. Implement this if there are special
        cleanup routines or sleeping until a certain moment is required.

        Args:
            uds: The UDSClient class where this hook is embedded. The caller typically
                 calls this function with `self` as the first argument.
            session: The desired session identifier.
        Returns:
            True on success, False on error.
        """
        return True

    async def check_and_set_session(
        self,
        expected_session: int,
        retries: int = 3,
    ) -> bool:
        """check_and_set_session() reads the current session and (re)tries to set
        the session to the expected session if they do not match.

        Returns True if the current session matches the expected session,
        or if read_session is not supported by the ECU or in the current session."""

        logger.debug(f"Checking current session, expecting {g_repr(expected_session)}")

        try:
            current_session = await self.read_session(config=UDSRequestConfig(max_retry=retries))
        except UnexpectedNegativeResponse as e:
            if suggests_identifier_not_supported(e.RESPONSE_CODE):
                logger.info(
                    f"Read current session not supported: {e.RESPONSE_CODE.name}, skipping check_session"
                )
                return True
            raise e
        except TimeoutError:
            logger.warning("Reading current session timed out, skipping check_session")
            return True

        logger.debug(f"Current session is {g_repr(current_session)}")
        if current_session == expected_session:
            return True

        for i in range(retries + 1):
            logger.warning(
                f"Not in session {g_repr(expected_session)}, ECU replied with {g_repr(current_session)}"
            )

            logger.info(
                f"Switching to session {g_repr(expected_session)}; attempt {i + 1} of {retries}"
            )
            resp = await self.set_session(expected_session)

            if isinstance(resp, service.NegativeResponse):
                logger.warning(f"Switching to session {g_repr(expected_session)} failed: {resp}")

            try:
                current_session = await self.read_session(
                    config=UDSRequestConfig(max_retry=retries)
                )
                logger.debug(f"Current session is {g_repr(current_session)}")
                if current_session == expected_session:
                    return True
            except UnexpectedNegativeResponse as e:
                if suggests_identifier_not_supported(e.RESPONSE_CODE):
                    logger.info(
                        f"Read current session not supported: {e.RESPONSE_CODE.name}, skipping check_session"
                    )
                    return True
                raise e
            except TimeoutError:
                logger.warning("Reading current session timed out, skipping check_session")
                return True

        logger.warning(
            f"Failed to switch to session {g_repr(expected_session)} after {retries} attempts"
        )
        return False

    async def power_cycle(self, sleep: float = 5) -> bool:
        if self.power_supply is None:
            logger.debug("no power_supply available")
            return False

        async def callback() -> None:
            await self.wait_for_ecu()

        await self.power_supply.power_cycle(sleep, callback)
        self.state.reset()
        return True

    async def leave_session(
        self,
        level: int,
        config: UDSRequestConfig | None = None,
        sleep: float | None = None,
    ) -> bool:
        """leave_session() is a hook which can be called explicitly by a
        scanner when a session is to be disabled. Use this hook if resetting
        the ECU is required, e.g. when disabling the programming session.
        """
        resp: service.UDSResponse = await self.ecu_reset(0x01)
        if isinstance(resp, service.NegativeResponse):
            if sleep is not None:
                await self.power_cycle(sleep=sleep)
            else:
                await self.power_cycle()
            await self.reconnect()
        await self.wait_for_ecu()

        resp = await self.set_session(0x01, config=config)
        if isinstance(resp, service.NegativeResponse):
            if sleep is not None:
                await self.power_cycle(sleep=sleep)
            else:
                await self.power_cycle()
            await self.reconnect()
        return True

    async def set_session(
        self,
        level: int,
        config: UDSRequestConfig | None = None,
        use_db: bool = True,
    ) -> service.NegativeResponse | service.DiagnosticSessionControlResponse:
        config = config if config is not None else UDSRequestConfig()

        if not config.skip_hooks:
            await self.set_session_pre(level, config=config)

        resp = await self.diagnostic_session_control(level, config=config)

        if isinstance(resp, service.NegativeResponse) and self.db_handler is not None and use_db:
            logger.debug("Could not switch to session. Trying with database transitions ...")

            if self.db_handler is not None:
                steps = await self.db_handler.get_session_transition(level)

                logger.debug(f"Found the following steps in database: {steps}")

                if steps is not None:
                    for step in steps:
                        await self.set_session(step, use_db=False)

                    resp = await self.diagnostic_session_control(level, config=config)

        if not isinstance(resp, service.NegativeResponse) and not config.skip_hooks:
            await self.set_session_post(level, config=config)

        return resp

    async def read_dtc(
        self, config: UDSRequestConfig | None = None
    ) -> service.NegativeResponse | service.ReportDTCByStatusMaskResponse:
        """Read all dtc records from the ecu."""
        return await self.read_dtc_information_report_dtc_by_status_mask(0xFF, config=config)

    async def clear_dtc(
        self, config: UDSRequestConfig | None = None
    ) -> service.NegativeResponse | service.ClearDiagnosticInformationResponse:
        """Clear all dtc records on the ecu."""
        return await self.clear_diagnostic_information(0xFFFFFF, config=config)

    async def read_vin(
        self, config: UDSRequestConfig | None = None
    ) -> service.NegativeResponse | service.ReadDataByIdentifierResponse:
        """Read the VIN of the vehicle"""
        return await self.read_data_by_identifier(0xF190, config=config)

    async def transmit_data(
        self,
        data: bytes,
        block_length: int,
        max_block_length: int = 0xFFF,
        config: UDSRequestConfig | None = None,
    ) -> None:
        """transmit_data splits the data to be sent in several blocks of size block_length,
        transfers all of them and concludes the transmission with RequestTransferExit"""
        if block_length > max_block_length:
            logger.warning(f"Limiting block size to {g_repr(max_block_length)}")
            block_length = max_block_length
        # block_length includes the service identifier and block counter; payload must be smaller
        payload_size = block_length - 2
        counter = 0
        for i in range(0, len(data), payload_size):
            counter += 1
            payload = data[i : i + payload_size]
            logger.debug(
                f"Transferring block {g_repr(counter)} with payload size {g_repr(len(payload))}"
            )
            resp: service.UDSResponse = await self.transfer_data(
                counter & 0xFF, payload, config=config
            )
            raise_for_error(resp, f"Transmitting data failed at index {g_repr(i)}")
        resp = await self.request_transfer_exit(config=config)
        raise_for_error(resp)

    async def _wait_for_ecu_endless_loop(self, sleep_time: float) -> None:
        """Internal method with endless loop in case of no answer from ECU"""
        config = UDSRequestConfig(timeout=0.5, max_retry=0, skip_hooks=True)
        i = -1
        while True:
            i = (i + 1) % 4
            logger.info(f"Waiting for ECU{'.' * i}")
            try:
                await asyncio.sleep(sleep_time)
                await self.ping(config=config)
                break
            # When the ECU is not ready, we expect an UDSException, e.g. MissingResponse.
            # On ConnectionError, we additionally reconnect the transport to ensure connectivity.
            # Since Gallia converts a ConnectionError to a MissingResponse in `request_unsafe`, however,
            # there is a need to reconnect the transport also in case of high-level UDSExceptions
            # such as MissingResponses that are raised (__cause__) from ConnectionErrors.
            except (ConnectionError, UDSException) as e:
                logger.debug(f"ECU not ready: {e!r}")
                if isinstance(e, ConnectionError) or isinstance(e.__cause__, ConnectionError):
                    logger.debug("Reconnecting…")
                    await self.reconnect()
        logger.info("ECU ready")

    async def wait_for_ecu(
        self,
        timeout: float | None = 10,
    ) -> bool:
        """Wait for ecu to be alive again (e.g. after reset).
        Sends a ping every 0.5s and waits at most timeout.
        If timeout is None, wait endlessly"""
        logger.info(f"Waiting for {timeout}s for ECU to respond")
        if self.tester_present_task and self.tester_present_interval:
            await self.stop_cyclic_tester_present()

        try:
            await asyncio.wait_for(self._wait_for_ecu_endless_loop(0.5), timeout=timeout)
            return True
        except TimeoutError:
            logger.critical("Timeout while waiting for ECU!")
            return False
        finally:
            if self.tester_present_task and self.tester_present_interval:
                await self.start_cyclic_tester_present(self.tester_present_interval)

    async def _tester_present_worker(self, interval: float) -> None:
        assert self.transport
        logger.debug("tester present worker started")
        task = asyncio.current_task()
        while task is not None and task.cancelling() == 0:
            try:
                await asyncio.sleep(interval)
                # TODO: Only ping if there was no other UDS traffic for `interval` amount of time
                await self.ping(UDSRequestConfig(max_retry=0))
            except asyncio.CancelledError:
                logger.debug("tester present worker terminated")
                raise
            except ConnectionError:
                logger.info("connection lost; tester present waiting…")
            except Exception as e:
                logger.warning(f"Tester present worker got {e!r}")
        logger.debug("Tester present worker was cancelled but received no asyncio.CancelledError")

    async def start_cyclic_tester_present(self, interval: float) -> None:
        logger.debug("Starting tester present worker")
        self.tester_present_interval = interval
        coroutine = self._tester_present_worker(interval)
        self.tester_present_task = asyncio.create_task(coroutine)
        self.tester_present_task.add_done_callback(
            handle_task_error,
            context=set_task_handler_ctx_variable(__name__, "TesterPresent"),
        )

        # enforce context switch
        # this ensures, that the task is executed at least once
        # if the task is not executed, task.cancel will fail with CancelledError
        await asyncio.sleep(0)

    async def stop_cyclic_tester_present(self) -> None:
        logger.debug("Stopping tester present worker")
        if self.tester_present_task is None:
            logger.warning("BUG: stop_cyclic_tester_present() called but no task running")
            return

        self.tester_present_task.cancel()
        try:
            await self.tester_present_task
        except asyncio.CancelledError:
            pass

    async def update_state(
        self, request: service.UDSRequest, response: service.UDSResponse
    ) -> None:
        if isinstance(response, service.DiagnosticSessionControlResponse):
            self.state.reset()
            self.state.session = response.diagnostic_session_type

        if (
            isinstance(response, service.ReadDataByIdentifierResponse)
            and response.data_identifier == DataIdentifier.ActiveDiagnosticSessionDataIdentifier
        ):
            new_session = int.from_bytes(response.data_record, "big")

            if self.state.session != new_session:
                self.state.reset()
                self.state.session = new_session

        if (
            isinstance(response, service.SecurityAccessResponse)
            and response.security_access_type % 2 == 0
        ):
            self.state.security_access_level = response.security_access_type - 1

        if isinstance(response, service.ECUResetResponse):
            self.state.reset()

    async def refresh_state(self, reset_state: bool = False) -> None:
        """
        Refresh the attributes of the ECU states, if possible.
        By, default, old values are only overwritten in case the corresponding
        information can be requested from the ECU and could be retrieved from a
        positive response from the ECU.

        :param reset_state: If True, the ECU state is reset before updating it.
        """
        if reset_state:
            self.state.reset()

        await self.read_session()

    async def _request(
        self, request: service.UDSRequest, config: UDSRequestConfig | None = None
    ) -> service.UDSResponse:
        """Sends a raw UDS request and returns the response.
        Network errors are handled via exponential backoff.
        Pending errors, triggered by the ECU are resolved as well.

        :param request: request to send
        :param config: The request config parameters
        :return: The response.
        """
        response = None
        exception: Exception | None = None
        send_time = datetime.now(UTC).astimezone()
        receive_time = None

        try:
            response = await super()._request(request, config)
            receive_time = datetime.now(UTC).astimezone()
            return response
        except ResponseException as e:
            exception = e
            response = e.response
            raise
        except Exception as e:
            exception = e
            raise
        finally:
            try:
                if self.implicit_logging and self.db_handler is not None:
                    mode = LogMode.implicit

                    if config is not None and config.tags is not None and "ANALYZE" in config.tags:
                        mode = LogMode.emphasized

                    await self.db_handler.insert_scan_result(
                        self.state.to_dict(),
                        service.UDSRequest.parse_dynamic(request.pdu),
                        response,
                        exception,
                        send_time,
                        receive_time,
                        mode,
                    )
            except Exception as e:
                logger.warning(f"Could not log messages to database: {g_repr(e)}")

            if response is not None:
                await self.update_state(request, response)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import binascii
import io
from abc import ABC, abstractmethod
from typing import Any, Protocol, Self
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from gallia.log import get_logger
from gallia.net import join_host_port
from gallia.transports.schemes import TransportScheme

logger = get_logger(__name__)


class TargetURI:
    """TargetURI represents a target to which gallia can connect.
    The target string must conform to a URI is specified by RFC3986.

    Basically, this is a wrapper around Python's ``urlparse()`` and
    ``parse_qs()`` methods. TargetURI provides frequently used properties
    for a more userfriendly usage. Instances are meant to be passed to
    :meth:`BaseTransport.connect()` of transport implementations.
    """

    def __init__(self, raw: str) -> None:
        self.raw = raw
        self.url = urlparse(raw)
        self.qs = parse_qs(self.url.query)

    @classmethod
    def from_parts(
        cls,
        scheme: str,
        host: str,
        port: int | None,
        args: dict[str, Any],
    ) -> Self:
        """Constructs a instance of TargetURI with the given arguments.
        The ``args`` dict is used for the query string.
        """
        netloc = host if port is None else join_host_port(host, port)
        return cls(urlunparse((scheme, netloc, "", "", urlencode(args), "")))

    @property
    def scheme(self) -> TransportScheme:
        """The URI scheme"""
        return TransportScheme(self.url.scheme)

    @property
    def hostname(self) -> str | None:
        """The hostname (without port)"""
        return self.url.hostname

    @property
    def port(self) -> int | None:
        """The port number"""
        return self.url.port

    @property
    def netloc(self) -> str:
        """The hostname and the portnumber, separated by a colon."""
        return self.url.netloc

    @property
    def path(self) -> str:
        """The path property of the url."""
        return self.url.path

    @property
    def location(self) -> str:
        """A URI string which only consists of the relevant scheme,
        the host and the port.
        """
        return f"{self.scheme}://{self.url.netloc}"

    @property
    def qs_flat(self) -> dict[str, str]:
        """A dict which contains the query string's key/value pairs.
        In case a key appears multiple times, this variant only
        contains the first found key/value pair. In contrast to
        :attr:`qs`, this variant avoids lists and might be easier
        to use for some cases.
        """
        d = {}
        for k, v in self.qs.items():
            d[k] = v[0]
        return d

    def __str__(self) -> str:
        return self.raw


class TransportProtocol(Protocol):
    mutex: asyncio.Lock
    target: TargetURI
    is_closed: bool

    def get_writer(self) -> asyncio.StreamWriter:
        raise NotImplementedError

    def get_reader(self) -> asyncio.StreamReader:
        raise NotImplementedError


class BaseTransport(ABC):
    """BaseTransport is the base class providing the required
    interface for all transports used by gallia.

    A transport usually is some kind of network protocol which
    carries an application level protocol. A good example is
    DoIP carrying UDS requests which acts as a minimal middleware
    on top of TCP.

    This class is to be used as a subclass with all abstractmethods
    implemented and the SCHEME property filled.

    A few methods provide a ``tags`` argument. The debug logs of these
    calls include these tags in the ``tags`` property of the relevant
    :class:`gallia.log.PenlogRecord`.
    """

    #: The scheme for the implemented protocol, e.g. "doip".
    SCHEME: str = ""
    #: The buffersize of the transport. Might be used in read() calls.
    #: Defaults to :const:`io.DEFAULT_BUFFER_SIZE`.
    BUFSIZE: int = io.DEFAULT_BUFFER_SIZE

    def __init__(self, target: TargetURI) -> None:
        self.mutex = asyncio.Lock()
        self.target = target
        self.is_closed = False

    def __init_subclass__(
        cls,
        /,
        scheme: str,
        bufsize: int = io.DEFAULT_BUFFER_SIZE,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        cls.SCHEME = scheme
        cls.BUFSIZE = bufsize

    @classmethod
    def check_scheme(cls, target: TargetURI) -> None:
        """Checks if the provided URI has the correct scheme."""
        if target.scheme != cls.SCHEME:
            raise ValueError(f"invalid scheme: {target.scheme}; expected: {cls.SCHEME}")

    @classmethod
    @abstractmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        """Classmethod to connect the transport to a relevant target.
        The target argument is a URI, such as `doip://192.0.2.2:13400?src_addr=0xf4&dst_addr=0x1d"`
        An instance of the relevant transport class is returned.
        """

    @abstractmethod
    async def close(self) -> None:
        """Terminates the connection and clean up all allocated resources."""

    async def reconnect(self, timeout: float | None = None) -> Self:
        """Closes the connection to the target and attempts to reconnect every
        100 ms until at max timeout. If timeout is None, only attempt to connect
        once.
        A new instance of this class is returned rendering the old one obsolete.
        This method is safe for concurrent use.
        """
        async with self.mutex:
            try:
                await self.close()
            except ConnectionError as e:
                logger.warning(f"close() failed during reconnect ({e!r}); ignoring")

            async with asyncio.timeout(timeout):
                logger.debug(
                    f"Attempting to establish a new connection with a timeout of {timeout}"
                )
                while True:
                    try:
                        logger.debug("Attempting to connect...")
                        return await self.connect(self.target)
                    except ConnectionError as e:
                        logger.info(f"Connection attempt failed while reconnecting: {e!r}")
                        if timeout is None:
                            logger.debug("Breaking out of the reconnect-loop since timeout is None")
                            raise e
                        await asyncio.sleep(0.1)

    @abstractmethod
    async def read(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        """Reads one message and returns its raw byte representation.
        An example for one message is 'one line, terminated by newline'
        for a TCP transport yielding lines.
        """

    @abstractmethod
    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        """Writes one message and return the number of written bytes."""

    async def request(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        """Chains a :meth:`write()` call with a :meth:`read()` call.
        The call is protected by a mutex and is thus safe for concurrent
        use.
        """
        async with self.mutex:
            return await self.request_unsafe(data, timeout, tags)

    async def request_unsafe(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        """Chains a :meth:`write()` call with a :meth:`read()` call.
        The call is **not** protected by a mutex. Only use this method
        when you know what you are doing.
        """
        await self.write(data, timeout, tags)
        return await self.read(timeout, tags)


class LinesTransportMixin:
    async def write(
        self: TransportProtocol,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        t = tags + ["write"] if tags is not None else ["write"]

        logger.trace(data.hex() + "0a", extra={"tags": t})

        writer = self.get_writer()
        writer.write(binascii.hexlify(data) + b"\n")
        await asyncio.wait_for(writer.drain(), timeout)
        return len(data)

    async def read(
        self: TransportProtocol,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        data = await asyncio.wait_for(self.get_reader().readline(), timeout)
        d = data.decode().strip()

        t = tags + ["read"] if tags is not None else ["read"]
        logger.trace(d + "0a", extra={"tags": t})

        return binascii.unhexlify(d)



# SPDX-FileCopyrightText: AISEC Pentesting Team
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import socket
import struct
from dataclasses import dataclass
from enum import IntEnum, unique
from typing import Any, Self

from pydantic import BaseModel, field_validator

from gallia.log import get_logger
from gallia.transports.base import BaseTransport, TargetURI
from gallia.utils import (
    auto_int,
    handle_task_error,
    set_task_handler_ctx_variable,
)

logger = get_logger(__name__)


@unique
class ProtocolVersions(IntEnum):
    ISO_13400_2_2010 = 0x01
    ISO_13400_2_2012 = 0x02
    ISO_13400_2_2019 = 0x03


@unique
class RoutingActivationRequestTypes(IntEnum):
    RESERVED = 0xFF
    ManufacturerSpecific = 0xFE
    Default = 0x00
    WWH_OBD = 0x01
    CentralSecurity = 0xE0

    @classmethod
    def _missing_(cls, value: Any) -> "RoutingActivationRequestTypes":
        if value in range(0xE1, 0x100):
            return cls.ManufacturerSpecific
        return cls.RESERVED


@unique
class RoutingActivationResponseCodes(IntEnum):
    RESERVED = 0xFF
    ManufacturerSpecific = 0xFE
    UnknownSourceAddress = 0x00
    NoResources = 0x01
    InvalidConnectionEntry = 0x02
    AlreadyActive = 0x03
    AuthenticationMissing = 0x04
    ConfirmationRejected = 0x05
    UnsupportedActivationType = 0x06
    TLSRequired = 0x07
    Success = 0x10
    SuccessConfirmationRequired = 0x11

    @classmethod
    def _missing_(cls, value: Any) -> "RoutingActivationResponseCodes":
        if value in range(0xE0, 0xFF):
            return cls.ManufacturerSpecific
        return cls.RESERVED


class DoIPRoutingActivationDeniedError(ConnectionAbortedError):
    rac_code: RoutingActivationResponseCodes

    def __init__(self, rac_code: int):
        self.rac_code = RoutingActivationResponseCodes(rac_code)
        super().__init__(f"DoIP routing activation denied: {self.rac_code.name} ({rac_code})")


@unique
class PayloadTypes(IntEnum):
    GenericDoIPHeaderNACK = 0x0000
    VehicleIdentificationRequestMessage = 0x0001
    VehicleIdentificationRequestMessageWithEID = 0x0002
    VehicleIdentificationRequestMessageWithVIN = 0x0003
    VehicleAnnouncementMessage = 0x004
    RoutingActivationRequest = 0x0005
    RoutingActivationResponse = 0x0006
    AliveCheckRequest = 0x0007
    AliveCheckResponse = 0x0008
    DoIPEntityStatusRequest = 0x4001
    DoIPEntityStatusResponse = 0x4002
    DiagnosticPowerModeInformationRequest = 0x4003
    DiagnosticPowerModeInformationResponse = 0x4004
    DiagnosticMessage = 0x8001
    DiagnosticMessagePositiveAcknowledgement = 0x8002
    DiagnosticMessageNegativeAcknowledgement = 0x8003


@unique
class DiagnosticMessagePositiveAckCodes(IntEnum):
    Success = 0x00


@unique
class DiagnosticMessageNegativeAckCodes(IntEnum):
    RESERVED = 0xFF
    InvalidSourceAddress = 0x02
    UnknownTargetAddress = 0x03
    DiagnosticMessageTooLarge = 0x04
    OutOfMemory = 0x05
    TargetUnreachable = 0x06
    UnknownNetwork = 0x07
    TransportProtocolError = 0x08

    @classmethod
    def _missing_(cls, value: Any) -> "DiagnosticMessageNegativeAckCodes":
        return cls.RESERVED


class DoIPNegativeAckError(BrokenPipeError):
    nack_code: DiagnosticMessageNegativeAckCodes

    def __init__(self, negative_ack_code: int):
        self.nack_code = DiagnosticMessageNegativeAckCodes(negative_ack_code)
        super().__init__(f"DoIP negative ACK received: {self.nack_code.name} ({negative_ack_code})")


@unique
class GenericDoIPHeaderNACKCodes(IntEnum):
    RESERVED = 0xFF
    IncorrectPatternFormat = 0x00
    UnknownPayloadType = 0x01
    MessageTooLarge = 0x02
    OutOfMemory = 0x03
    InvalidPayloadLength = 0x04

    @classmethod
    def _missing_(cls, value: Any) -> "GenericDoIPHeaderNACKCodes":
        return cls.RESERVED


class DoIPGenericHeaderNACKError(ConnectionAbortedError):
    nack_code: GenericDoIPHeaderNACKCodes

    def __init__(self, nack_code: int):
        self.nack_code = GenericDoIPHeaderNACKCodes(nack_code)
        super().__init__(f"DoIP generic header negative ACK: {self.nack_code.name} ({nack_code})")


class TimingAndCommunicationParameters(IntEnum):
    CtrlTimeout = 2000
    AnnounceWait = 500
    AnnounceInterval = 500
    AnnounceNum = 3
    DiagnosticMessageMessageAckTimeout = 2000
    RoutingActivationResponseTimeout = 2000
    DiagnosticMessageMessageTimeout = 2000
    TCPGeneralInactivityTimeout = 5000
    TCPInitialInactivityTimeout = 2000
    TCPAliveCheckTimeout = 500
    ProcessingTimeout = 2000
    VehicleDiscoveryTimeout = 5000


@dataclass
class GenericHeader:
    ProtocolVersion: int
    PayloadType: int
    PayloadLength: int

    def pack(self) -> bytes:
        return struct.pack(
            "!BBHL",
            self.ProtocolVersion,
            self.ProtocolVersion ^ 0xFF,
            self.PayloadType,
            self.PayloadLength,
        )

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        (
            protocol_version,
            inverse_protocol_version,
            payload_type,
            payload_length,
        ) = struct.unpack("!BBHL", data)
        if protocol_version != inverse_protocol_version ^ 0xFF:
            raise ValueError("inverse protocol_version is invalid")
        return cls(
            protocol_version,
            payload_type,
            payload_length,
        )


@dataclass
class GenericDoIPHeaderNACK:
    GenericHeaderNACKCode: GenericDoIPHeaderNACKCodes

    def pack(self) -> bytes:
        return struct.pack(
            "!B",
            self.GenericHeaderNACKCode,
        )

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        (generic_header_NACK_code,) = struct.unpack("!B", data)
        return cls(
            GenericDoIPHeaderNACKCodes(generic_header_NACK_code),
        )


@dataclass
class VehicleIdentificationRequestMessage:
    def pack(self) -> bytes:
        return b""


@unique
class FurtherActionCodes(IntEnum):
    RESERVED = 0x0F
    ManufacturerSpecific = 0xFF
    NoFurtherActionRequired = 0x00
    RoutingActivationRequiredToInitiateCentralSecurity = 0x10

    @classmethod
    def _missing_(cls, value: Any) -> "FurtherActionCodes":
        if value in range(0x11, 0x100):
            return cls.ManufacturerSpecific
        return cls.RESERVED


@unique
class SynchronisationStatusCodes(IntEnum):
    RESERVED = 0xFF
    VINGIDSynchronized = 0x00
    IncompleteVINGIDNotSynchronized = 0x10

    @classmethod
    def _missing_(cls, value: Any) -> "SynchronisationStatusCodes":
        return cls.RESERVED


@dataclass
class VehicleAnnouncementMessage:
    VIN: bytes
    LogicalAddress: int
    EID: bytes
    GID: bytes
    FurtherActionRequired: FurtherActionCodes
    VINGIDSyncStatus: SynchronisationStatusCodes | None

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        if len(data) == 32:
            # VINGIDSyncStatus is optional
            (vin, logical_address, eid, gid, further_action_required) = struct.unpack(
                "!17sH6s6sB", data
            )
            vin_gid_sync_status = None
        else:
            (
                vin,
                logical_address,
                eid,
                gid,
                further_action_required,
                vin_gid_sync_status,
            ) = struct.unpack("!17sH6s6sBB", data)

        return cls(
            vin,
            logical_address,
            eid,
            gid,
            FurtherActionCodes(further_action_required),
            SynchronisationStatusCodes(vin_gid_sync_status)
            if vin_gid_sync_status is not None
            else None,
        )


@dataclass
class DoIPEntityStatusRequest:
    def pack(self) -> bytes:
        return b""


@unique
class NodeTypes(IntEnum):
    RESERVED = 0xFF
    Gateway = 0x00
    Node = 0x01

    @classmethod
    def _missing_(cls, value: Any) -> "NodeTypes":
        return cls.RESERVED


@dataclass
class DoIPEntityStatusResponse:
    NodeType: NodeTypes
    MaximumConcurrentTCP_DATASockets: int
    CurrentlyOpenTCP_DATASockets: int
    MaximumDataSize: int | None

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        if len(data) == 3:
            # MaximumDataSize is optional
            (nt, mcts, ncts) = struct.unpack("!BBB", data)
            mds = None
        else:
            (nt, mcts, ncts, mds) = struct.unpack("!BBBI", data)

        return cls(NodeTypes(nt), mcts, ncts, mds)


@dataclass
class RoutingActivationRequest:
    SourceAddress: int
    ActivationType: int
    Reserved: int = 0x00000000  # Not used, default value.
    # OEMReserved uint32

    def pack(self) -> bytes:
        return struct.pack("!HBI", self.SourceAddress, self.ActivationType, self.Reserved)


@dataclass
class RoutingActivationResponse:
    SourceAddress: int
    TargetAddress: int
    RoutingActivationResponseCode: int
    Reserved: int = 0x00000000  # Not used, default value.
    # OEMReserved uint32

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        (
            source_address,
            target_address,
            routing_activation_response_code,
            reserved,
        ) = struct.unpack("!HHBI", data)
        if reserved != 0x00000000:
            raise ValueError("reserved field contains data")
        return cls(
            source_address,
            target_address,
            routing_activation_response_code,
            reserved,
        )


@dataclass
class DiagnosticMessage:
    SourceAddress: int
    TargetAddress: int
    UserData: bytes

    def pack(self) -> bytes:
        return (
            struct.pack(
                "!HH",
                self.SourceAddress,
                self.TargetAddress,
            )
            + self.UserData
        )

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        source_address, target_address = struct.unpack("!HH", data[:4])
        data = data[4:]
        return cls(source_address, target_address, data)


@dataclass
class DiagnosticMessageAcknowledgement:
    SourceAddress: int
    TargetAddress: int
    ACKCode: int
    PreviousDiagnosticMessageData: bytes

    def pack(self) -> bytes:
        return (
            struct.pack(
                "!HHB",
                self.SourceAddress,
                self.TargetAddress,
                self.ACKCode,
            )
            + self.PreviousDiagnosticMessageData
        )


class DiagnosticMessagePositiveAcknowledgement(DiagnosticMessageAcknowledgement):
    ACKCode: DiagnosticMessagePositiveAckCodes

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        source_address, target_address, ack_code = struct.unpack("!HHB", data[:5])
        prev_data = data[5:]

        return cls(
            source_address,
            target_address,
            DiagnosticMessagePositiveAckCodes(ack_code),
            prev_data,
        )


class DiagnosticMessageNegativeAcknowledgement(DiagnosticMessageAcknowledgement):
    ACKCode: DiagnosticMessageNegativeAckCodes

    @classmethod
    def unpack(cls, data: bytes) -> Self:
        source_address, target_address, ack_code = struct.unpack("!HHB", data[:5])
        prev_data = data[5:]

        return cls(
            source_address,
            target_address,
            DiagnosticMessageNegativeAckCodes(ack_code),
            prev_data,
        )


@dataclass
class AliveCheckRequest:
    pass


@dataclass
class AliveCheckResponse:
    SourceAddress: int

    def pack(self) -> bytes:
        return struct.pack("!H", self.SourceAddress)


# Messages expected to be sent by the DoIP gateway.
DoIPInData = (
    GenericDoIPHeaderNACK
    | RoutingActivationResponse
    | DiagnosticMessage
    | DiagnosticMessagePositiveAcknowledgement
    | DiagnosticMessageNegativeAcknowledgement
    | AliveCheckRequest
)

# Messages expected to be sent by us.
DoIPOutData = RoutingActivationRequest | DiagnosticMessage | AliveCheckResponse

DoIPFrame = tuple[
    GenericHeader,
    DoIPInData | DoIPOutData,
]
DoIPDiagFrame = tuple[GenericHeader, DiagnosticMessage]


class DoIPConnection:
    def __init__(  # noqa: PLR0913
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        src_addr: int,
        target_addr: int,
        protocol_version: int,
        separate_diagnostic_message_queue: bool = False,
    ):
        self.reader = reader
        self.writer = writer
        self.src_addr = src_addr
        self.target_addr = target_addr
        self.protocol_version = protocol_version
        self.separate_diagnostic_message_queue = separate_diagnostic_message_queue
        self._diagnostic_message_queue: asyncio.Queue[DoIPDiagFrame] = asyncio.Queue()
        self._read_queue: asyncio.Queue[DoIPFrame] = asyncio.Queue()
        self._read_task = asyncio.create_task(self._read_worker())
        self._read_task.add_done_callback(
            handle_task_error,
            context=set_task_handler_ctx_variable(__name__, "DoipReader"),
        )
        self._is_closed = False
        self._mutex = asyncio.Lock()

    @classmethod
    async def connect(  # noqa: PLR0913
        cls,
        host: str,
        port: int,
        src_addr: int,
        target_addr: int,
        so_linger: bool = False,
        protocol_version: int = ProtocolVersions.ISO_13400_2_2019,
        separate_diagnostic_message_queue: bool = False,
    ) -> Self:
        reader, writer = await asyncio.open_connection(host, port)

        if so_linger is True:
            # Depending on who will close the connection in the end, one party's socket
            # will remain in a TIME_WAIT state, which occupies resources until enough
            # time has passed. Setting the LINGER socket option tells our kernel to
            # close the connection with a RST, which brings the TCP connection to an
            # error state and thus avoids TIME_WAIT and instantly forces LISTEN or CLOSED
            # For more info, see e.g. Note 3 of :
            # https://www.ietf.org/rfc/rfc9293.html#name-state-machine-overview
            sock = writer.get_extra_info("socket")
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack("ii", 1, 0))

        return cls(
            reader,
            writer,
            src_addr,
            target_addr,
            protocol_version,
            separate_diagnostic_message_queue,
        )

    async def _read_frame(self) -> DoIPFrame | tuple[None, None]:
        # Header is fixed size 8 byte.
        hdr_buf = await self.reader.readexactly(8)
        hdr = GenericHeader.unpack(hdr_buf)

        payload_buf = await self.reader.readexactly(hdr.PayloadLength)
        payload: DoIPInData
        match hdr.PayloadType:
            case PayloadTypes.GenericDoIPHeaderNACK:
                payload = GenericDoIPHeaderNACK.unpack(payload_buf)
            case PayloadTypes.RoutingActivationResponse:
                payload = RoutingActivationResponse.unpack(payload_buf)
            case PayloadTypes.DiagnosticMessagePositiveAcknowledgement:
                payload = DiagnosticMessagePositiveAcknowledgement.unpack(payload_buf)
            case PayloadTypes.DiagnosticMessageNegativeAcknowledgement:
                payload = DiagnosticMessageNegativeAcknowledgement.unpack(payload_buf)
            case PayloadTypes.DiagnosticMessage:
                payload = DiagnosticMessage.unpack(payload_buf)
            case PayloadTypes.AliveCheckRequest:
                payload = AliveCheckRequest()
            case _:
                logger.warning(
                    f"DoIP message with unhandled PayloadType: {hdr} {payload_buf.hex()}"
                )
                return None, None
        logger.trace("Received DoIP message: %s, %s", hdr, payload)
        return hdr, payload

    async def _read_worker(self) -> None:
        try:
            while True:
                hdr, data = await self._read_frame()
                if hdr is None or data is None:
                    continue
                if hdr.PayloadType == PayloadTypes.AliveCheckRequest:
                    await self.write_alive_check_response()
                    continue
                if isinstance(data, DiagnosticMessage) and self.separate_diagnostic_message_queue:
                    await self._diagnostic_message_queue.put((hdr, data))
                    continue
                await self._read_queue.put((hdr, data))
        except asyncio.CancelledError:
            logger.debug("DoIP read worker got cancelled")
        except asyncio.IncompleteReadError as e:
            logger.debug(f"DoIP read worker received EOF: {e!r}")
        except Exception as e:
            logger.info(f"DoIP read worker died with {e!r}")
        finally:
            logger.debug("Feeding EOF to reader and requesting a close")
            self.reader.feed_eof()
            await self.close()

    async def read_frame_unsafe(self) -> DoIPFrame:
        # Avoid waiting on the queue forever when
        # the connection has been terminated.
        if self._is_closed:
            raise ConnectionError
        return await self._read_queue.get()

    async def read_frame(self) -> DoIPFrame:
        async with self._mutex:
            return await self.read_frame_unsafe()

    async def read_diag_request_raw(self) -> DoIPDiagFrame:
        unexpected_packets: list[tuple[Any, Any]] = []
        while True:
            if self.separate_diagnostic_message_queue:
                return await self._diagnostic_message_queue.get()
            hdr, payload = await self.read_frame()
            if not isinstance(payload, DiagnosticMessage):
                logger.warning(f"expected DoIP DiagnosticMessage, instead got: {hdr} {payload}")
                unexpected_packets.append((hdr, payload))
                continue
            if payload.SourceAddress != self.target_addr or payload.TargetAddress != self.src_addr:
                logger.warning(
                    f"DoIP-DiagnosticMessage: unexpected addresses (src:dst); expected {self.target_addr:#04x}:"
                    + f"{self.src_addr:#04x} but got: {payload.SourceAddress:#04x}:{payload.TargetAddress:#04x}"
                )
                unexpected_packets.append((hdr, payload))
                continue

            # Do not consume unexpected packets, but re-add them to the queue for other consumers
            for item in unexpected_packets:
                await self._read_queue.put(item)

            return hdr, payload

    async def read_diag_request(self) -> bytes:
        _, payload = await self.read_diag_request_raw()
        return payload.UserData

    async def _read_ack(self, prev_data: bytes) -> None:
        """This function consumes all DoIP Diagnostic Message ACKs and just logs a warning in case of mismatches."""

        unexpected_packets: list[tuple[Any, Any]] = []
        while True:
            hdr, payload = await self.read_frame_unsafe()
            if not isinstance(payload, DiagnosticMessagePositiveAcknowledgement) and not isinstance(
                payload, DiagnosticMessageNegativeAcknowledgement
            ):
                logger.warning(f"expected DoIP positive/negative ACK, instead got: {hdr} {payload}")
                unexpected_packets.append((hdr, payload))
                continue

            if payload.SourceAddress != self.target_addr or payload.TargetAddress != self.src_addr:
                logger.warning(
                    f"DoIP-ACK: unexpected addresses (src:dst); expected {self.target_addr:#04x}:{self.src_addr:#04x} "
                    + f"but got: {payload.SourceAddress:#04x}:{payload.TargetAddress:#04x}"
                )
                continue
            if (
                len(payload.PreviousDiagnosticMessageData) > 0
                and payload.PreviousDiagnosticMessageData
                != prev_data[: len(payload.PreviousDiagnosticMessageData)]
            ):
                logger.warning(
                    f"DoIP-ACK: got: {payload.PreviousDiagnosticMessageData.hex()}; expected: {prev_data.hex()}"
                )
                continue

            # Do not consume unexpected packets, but re-add them to the queue for other consumers
            # Mismatched DoIP ACKs, however, are not re-added to avoid unnecessary confusion
            for item in unexpected_packets:
                await self._read_queue.put(item)

            if isinstance(payload, DiagnosticMessageNegativeAcknowledgement):
                raise DoIPNegativeAckError(payload.ACKCode)
            return

    async def _read_routing_activation_response(self) -> None:
        unexpected_packets: list[tuple[Any, Any]] = []
        while True:
            hdr, payload = await self.read_frame_unsafe()
            if not isinstance(payload, RoutingActivationResponse):
                logger.warning(
                    f"expected DoIP RoutingActivationResponse, instead got: {hdr} {payload}"
                )
                unexpected_packets.append((hdr, payload))
                continue

            # Do not consume unexpected packets, but re-add them to the queue for other consumers
            for item in unexpected_packets:
                await self._read_queue.put(item)

            if payload.RoutingActivationResponseCode != RoutingActivationResponseCodes.Success:
                raise DoIPRoutingActivationDeniedError(payload.RoutingActivationResponseCode)
            return

    async def write_request_raw(self, hdr: GenericHeader, payload: DoIPOutData) -> None:
        async with self._mutex:
            buf = b""
            buf += hdr.pack()
            buf += payload.pack()
            self.writer.write(buf)
            await self.writer.drain()

            logger.trace("Sent DoIP message: hdr: %s, payload: %s", hdr, payload)

    async def write_diag_request(self, data: bytes) -> None:
        hdr = GenericHeader(
            ProtocolVersion=self.protocol_version,
            PayloadType=PayloadTypes.DiagnosticMessage,
            PayloadLength=len(data) + 4,
        )
        payload = DiagnosticMessage(
            SourceAddress=self.src_addr,
            TargetAddress=self.target_addr,
            UserData=data,
        )
        await self.write_request_raw(hdr, payload)

        # DiagnosticMessages will trigger an ACK message of the DoIP node/gateway that needs to be handled
        try:
            await asyncio.wait_for(
                self._read_ack(payload.UserData),
                TimingAndCommunicationParameters.DiagnosticMessageMessageAckTimeout / 1000,
            )
        except TimeoutError:
            logger.warning(f"Timeout while waiting for DoIP ACK for diagnostic message {payload}")
            # While not receiving a DoIP ACK is a violation of the standard, we deliberately
            # do not close the connection or raise a BrokenPipeError here, since it's not yet a problem

    async def write_routing_activation_request(
        self,
        activation_type: int,
    ) -> None:
        hdr = GenericHeader(
            ProtocolVersion=self.protocol_version,
            PayloadType=PayloadTypes.RoutingActivationRequest,
            PayloadLength=7,
        )
        payload = RoutingActivationRequest(
            SourceAddress=self.src_addr,
            ActivationType=activation_type,
            Reserved=0x00,
        )
        await self.write_request_raw(hdr, payload)

        # Sending a routing activation request triggers an immediate reply
        try:
            await asyncio.wait_for(
                self._read_routing_activation_response(),
                TimingAndCommunicationParameters.RoutingActivationResponseTimeout / 1000,
            )
        except TimeoutError as e:
            await self.close()
            raise BrokenPipeError("Timeout while waiting for Routing Activation Response") from e

    async def write_alive_check_response(self) -> None:
        hdr = GenericHeader(
            ProtocolVersion=self.protocol_version,
            PayloadType=PayloadTypes.AliveCheckResponse,
            PayloadLength=2,
        )
        payload = AliveCheckResponse(
            SourceAddress=self.src_addr,
        )
        await self.write_request_raw(hdr, payload)

    async def close(self) -> None:
        logger.debug("Closing DoIP connection...")
        if self._is_closed:
            logger.debug("DoIP connection already closed!")
            return
        self._is_closed = True
        logger.debug("Cancelling read worker")
        self._read_task.cancel()
        self.writer.close()
        logger.debug("Awaiting confirmation of closed writer")
        try:
            await self.writer.wait_closed()
        except ConnectionError as e:
            logger.debug(f"Exception while waiting for the writer to close: {e!r}")


class DoIPConfig(BaseModel):
    src_addr: int
    target_addr: int
    activation_type: int = RoutingActivationRequestTypes.WWH_OBD.value
    protocol_version: int = ProtocolVersions.ISO_13400_2_2019

    @field_validator(
        "src_addr",
        "target_addr",
        "activation_type",
        "protocol_version",
        mode="before",
    )
    def auto_int(cls, v: str) -> int:
        return auto_int(v)


class DoIPTransport(BaseTransport, scheme="doip"):
    def __init__(
        self,
        target: TargetURI,
        port: int,
        config: DoIPConfig,
        conn: DoIPConnection,
    ):
        super().__init__(target)
        self.port = port
        self.config = config
        self._conn = conn
        self._is_closed = False

    @staticmethod
    async def _connect(  # noqa: PLR0913
        hostname: str,
        port: int,
        src_addr: int,
        target_addr: int,
        activation_type: int,
        protocol_version: int,
    ) -> DoIPConnection:
        conn = await DoIPConnection.connect(
            hostname,
            port,
            src_addr,
            target_addr,
            protocol_version=protocol_version,
        )
        await conn.write_routing_activation_request(RoutingActivationRequestTypes(activation_type))
        return conn

    @classmethod
    async def connect(
        cls,
        target: str | TargetURI,
        timeout: float | None = None,
    ) -> Self:
        t = target if isinstance(target, TargetURI) else TargetURI(target)
        cls.check_scheme(t)

        if t.hostname is None:
            raise ValueError("no hostname specified")

        port = t.port if t.port is not None else 13400
        config = DoIPConfig.model_validate(t.qs_flat)
        conn = await asyncio.wait_for(
            cls._connect(
                t.hostname,
                port,
                config.src_addr,
                config.target_addr,
                config.activation_type,
                config.protocol_version,
            ),
            timeout,
        )
        return cls(t, port, config, conn)

    async def reconnect(self, timeout: float | None = None) -> Self:
        # It might be that the DoIP endpoint is not immediately ready for another
        # connection, so set the timeout to 10s by default.
        return await super().reconnect(10 if timeout is None else timeout)

    async def close(self) -> None:
        logger.debug("Closing DoIP transport...")
        if self._is_closed:
            logger.debug("DoIP transport already closed")
            return
        self._is_closed = True
        await self._conn.close()

    async def read(
        self,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> bytes:
        data = await asyncio.wait_for(self._conn.read_diag_request(), timeout)

        t = tags + ["read"] if tags is not None else ["read"]
        logger.trace(data.hex(), extra={"tags": t})
        return data

    async def write(
        self,
        data: bytes,
        timeout: float | None = None,
        tags: list[str] | None = None,
    ) -> int:
        t = tags + ["write"] if tags is not None else ["write"]
        logger.trace(data.hex(), extra={"tags": t})

        try:
            await asyncio.wait_for(self._conn.write_diag_request(data), timeout)
        except DoIPNegativeAckError as e:
            if e.nack_code != DiagnosticMessageNegativeAckCodes.TargetUnreachable:
                raise e
            # TargetUnreachable can be just a temporary issue. Thus, we do not raise
            # BrokenPipeError but instead ignore it here and let upper layers handle
            # missing responses
            logger.debug("DoIP message was ACKed with TargetUnreachable")

        return len(data)



# SPDX-FileCopyrightText: Hayden Richards
#
# SPDX-License-Identifier: MIT

"""Parses Pydantic Fields to Command-Line Arguments.

This package contains the functions required for parsing `pydantic` model
fields to `ArgumentParser` command-line arguments.

The public interface exposed by this package is the `parsing` modules, which
each contain the `should_parse()` and `parse_field()` functions.
"""

from gallia.pydantic_argparse.utils.pydantic import PydanticField

from . import (
    boolean,
    container,
    enum,
    literal,
    standard,
)
from .utils import SupportsAddArgument


def add_field(
    parser: SupportsAddArgument,
    field: PydanticField,
) -> None:
    """Parses pydantic field type, and then adds it to argument parser.

    Args:
        parser (argparse.ArgumentParser | argparse._ArgumentGroup): Sub-parser to add to.
        field (pydantic.fields.ModelField): Field to be added to parser.
    """
    # Switch on Field Type -- for fields that are pydantic models
    # this gets handled at the top level to distinguish
    # subcommands from arg groups
    if boolean.should_parse(field):
        boolean.parse_field(parser, field)
    elif container.should_parse(field):
        container.parse_field(parser, field)
    elif literal.should_parse(field):
        literal.parse_field(parser, field)
    elif enum.should_parse(field):
        enum.parse_field(parser, field)
    else:
        standard.parse_field(parser, field)
